"""
This module is designed to be imported at Maya startup, and should be added
to your userSetup.py script by running `mayapy -m maya_zen_tools.install`.
"""

from __future__ import annotations

import contextlib

from maya import cmds  # type: ignore

from maya_zen_tools.menu import create_menu
from maya_zen_tools.upgrade import upgrade


def set_selection_priority() -> None:
    """
    This sets selection priority needed for manipulating deformers effectively
    """
    surface_selection_priority: int = (
        cmds.selectPriority(
            query=True,
            nurbsSurface=True,
        )
        or 0
    )
    curve_selection_priority: int = (
        cmds.selectPriority(
            query=True,
            nurbsCurve=True,
        )
        or 0
    )
    polymesh_selection_priority: int = (
        cmds.selectPriority(
            query=True,
            polymesh=True,
        )
        or 0
    )
    locator_selection_priority: int = (
        cmds.selectPriority(
            query=True,
            locatorXYZ=True,
        )
        or 0
    )
    # Surface selection should be higher priority than polymesh selection
    if surface_selection_priority <= polymesh_selection_priority:
        surface_selection_priority = polymesh_selection_priority + 1
        cmds.selectPriority(
            nurbsSurface=surface_selection_priority,
        )
    # Curve selection should be higher priority than surface selection
    if curve_selection_priority <= surface_selection_priority:
        curve_selection_priority = surface_selection_priority + 1
        cmds.selectPriority(
            nurbsCurve=curve_selection_priority,
        )
    # Locator selection priority should be higher than curve selection
    if locator_selection_priority <= curve_selection_priority:
        locator_selection_priority = curve_selection_priority + 1
        cmds.selectPriority(
            locatorXYZ=locator_selection_priority,
        )


def main() -> None:
    """
    The main entry point for `maya-zen-tools startup`.
    """
    # Don't raise errors if the upgrade fails, just continue to use the
    # installed version
    with contextlib.suppress(Exception):
        upgrade()
    # Set selection preferences to track selection order
    cmds.selectPref(
        trackSelectionOrder=True,
    )
    set_selection_priority()
    create_menu()


cmds.evalDeferred(main)



"""
This module modifies your userSetup.py script to add startup procedures
needed to use ZenTools.
"""

from __future__ import annotations

import argparse
import re
from subprocess import check_call
from typing import TYPE_CHECKING

from maya import cmds  # type: ignore

from maya_zen_tools._utilities import find_user_setup_py, which_mayapy
from maya_zen_tools.menu import MENU

if TYPE_CHECKING:
    from pathlib import Path


def uninstall() -> None:
    """
    Uninstall ZenTools for Maya
    """
    user_setup_py: str = ""
    user_setup_py_path: Path = find_user_setup_py()
    if user_setup_py_path.is_file():
        with open(user_setup_py_path) as user_setup_py_io:
            user_setup_py = user_setup_py_io.read()
    # Remove the ZenTools startup line from userSetup.py
    if user_setup_py:
        with open(user_setup_py_path, "w") as user_setup_py_io:
            user_setup_py_io.write(
                re.sub(
                    r"(^|\n)from maya_zen_tools import startup(\n|$)",
                    r"\1",
                    user_setup_py,
                )
            )
    # Uninstall the `maya-zen-tools` package
    check_call(
        [
            str(which_mayapy()),
            "-m",
            "pip",
            "uninstall",
            "-y",
            "maya-zen-tools",
        ]
    )
    # Delete the ZenTools menu
    cmds.deleteUI(MENU)


def main() -> None:
    parser: argparse.ArgumentParser = argparse.ArgumentParser(
        prog="maya-zen-tools uninstall",
        description="Uninstall ZenTools for Maya",
    )
    parser.parse_args()
    uninstall()


if __name__ == "__main__":
    main()



from __future__ import annotations

from typing import Sequence


class ZenToolsWarning(Warning):
    pass


class MultipleVertexPathsPossibleWarning(ZenToolsWarning):
    pass


class ZenToolsError(Exception):
    """
    Base class for ZenTools Exceptions
    """


class InvalidSelectionError(Exception):
    pass


class EdgesNotOnSameRingError(InvalidSelectionError):
    def __init__(self, shape: str, edge_ids: tuple[int, int]) -> None:
        self.shape: str = shape
        self.edge_ids: tuple[int, ...] = edge_ids
        super().__init__(
            "Edges are not on the same ring: "
            f'("{shape}.e[{edge_ids[0]}]", "{shape}.e[{edge_ids[1]}]")'
        )


class NonLinearSelectionError(InvalidSelectionError):
    pass


class NonContiguousMeshSelectionError(InvalidSelectionError):
    pass


class TooManyShapesError(InvalidSelectionError):
    """
    Raised when a loop or loft is attempted using
    components from more than one polygon mesh.
    """

    def __init__(self, shapes: Sequence[str]) -> None:
        self.shapes: tuple[str, ...] = tuple(shapes)
        super().__init__(shapes)

    def __repr__(self) -> str:
        return (
            f"TooManyShapesError({self.shapes!r}): "
            "Selected components must all belong to the same shape."
        )

    def __str__(self) -> str:
        return repr(self)



from __future__ import annotations

from typing import Iterable

from maya import cmds  # type: ignore

from maya_zen_tools._traverse import (
    add_shared_face_edge_uvs,
    get_components_shape,
    get_shared_edge_vertices,
    iter_selected_components,
)


def _iter_flood_select_vertices(
    selected_vertices: Iterable[str], selected_edges: Iterable[str]
) -> Iterable[str]:
    if not selected_vertices:
        return
    border_vertices: set[str] = (
        set(
            cmds.ls(
                *cmds.polyListComponentConversion(
                    *selected_edges, fromEdge=True, toVertex=True
                ),
                flatten=True,
            )
        )
        if selected_edges
        else set()
    )
    vertices: set[str] = set(selected_vertices)
    add_vertices: set[str] = set(selected_vertices)
    # Expand the vertex selection until no new vertices are added
    while add_vertices:
        yield from add_vertices
        # By removing border vertices, we ensure that our traversal will stop
        # at the border, and once everything inside the border has been
        # added/yielded, `add_vertices` will be empty, thereby halting the
        # loop
        add_vertices = get_shared_edge_vertices(vertices) - border_vertices
        vertices |= add_vertices
    yield from border_vertices


def _iter_flood_select_faces(
    selected_faces: Iterable[str], selected_edges: Iterable[str]
) -> Iterable[str]:
    if not selected_faces:
        return
    yield from set(
        cmds.ls(
            *cmds.polyListComponentConversion(
                *_iter_flood_select_vertices(
                    set(
                        cmds.ls(
                            *cmds.polyListComponentConversion(
                                *selected_faces,
                                fromFace=True,
                                toVertex=True,
                            ),
                            flatten=True,
                        )
                    ),
                    selected_edges,
                ),
                fromVertex=True,
                toFace=True,
                internal=True,
            ),
            flatten=True,
        )
    )


def _iter_flood_select_uvs(
    selected_uvs: Iterable[str], selected_edges: Iterable[str]
) -> Iterable[str]:
    if not selected_uvs:
        return
    border_uvs: set[str] = (
        set(
            cmds.ls(
                *cmds.polyListComponentConversion(
                    *selected_edges, fromEdge=True, toVertex=True
                ),
                flatten=True,
            )
        )
        if selected_edges
        else set()
    )
    uvs: set[str] = set(selected_uvs)
    add_uvs: set[str] = set(selected_uvs)
    # Expand the vertex selection until no new uvs are added
    while add_uvs:
        yield from add_uvs
        # By removing border uvs, we ensure that our traversal will stop
        # at the border, and once everything inside the border has been
        # added/yielded, `add_uvs` will be empty, thereby halting the
        # loop
        add_uvs = add_shared_face_edge_uvs(uvs) - border_uvs
        uvs |= add_uvs
    yield from border_uvs


def flood_select(*selection: str) -> tuple[str, ...]:
    """
    Given a `selection` comprised of:

    - One or more polymesh faces, vertices, or UVs, and...
    - A set of edges enclosing an area around the faces, vertices, or
      UVs

    ...this function will expand the face, vertex, or UV selection to
    encompass the area enclosed by the selected edges.
    """
    cmds.waitCursor(state=True)
    try:
        selection = selection or tuple(cmds.ls(selection=True, flatten=True))
        selected_faces: tuple[str, ...] = tuple(
            iter_selected_components("f", selection=selection)
        )
        selected_vertices: tuple[str, ...] = tuple(
            iter_selected_components("vtx", selection=selection)
        )
        selected_uvs: tuple[str, ...] = tuple(
            iter_selected_components("map", selection=selection)
        )
        selected_edges: tuple[str, ...] = tuple(
            iter_selected_components("e", selection=selection)
        )
        # Raise an error if selected vertices span more than one mesh
        get_components_shape(selected_faces + selected_vertices + selected_uvs)
        selected_components: tuple[str, ...] = (
            tuple(
                _iter_flood_select_vertices(selected_vertices, selected_edges)
            )
            + tuple(_iter_flood_select_uvs(selected_uvs, selected_edges))
            + tuple(_iter_flood_select_faces(selected_faces, selected_edges))
        )
        cmds.select(
            *selection,
            deselect=True,
        )
        cmds.select(*selected_components, add=True)
    finally:
        if cmds.waitCursor(query=True, state=True):
            cmds.waitCursor(state=False)
    return selected_components



from __future__ import annotations

import contextlib
from datetime import datetime, timezone

from maya import cmds  # type: ignore

from maya_zen_tools._utilities import (
    get_maya_zen_tools_package_info,  # type: ignore
)
from maya_zen_tools.options import get_tool_option

# UI Components
MAYA_WINDOW: str = "MayaWindow"
MENU_SET: str = "modelingMenuSet"
MENU: str = "zenToolsMenu"

# Labels
SELECT_EDGES_BETWEEN_VERTICES_LABEL: str = "Select Edges Between Vertices"
SELECT_EDGES_BETWEEN_UVS_LABEL: str = "Select Edges Between UVs"
SELECT_UVS_BETWEEN_UVS_LABEL: str = "Select UVs Between UVs"
FLOOD_SELECT_LABEL: str = "Flood Select"
CURVE_DISTRIBUTE_BETWEEN_VERTICES_LABEL: str = (
    "Curve Distribute Between Vertices"
)
LOFT_DISTRIBUTE_VERTICES_BETWEEN_EDGES_LABEL: str = (
    "Loft Distribute Vertices Between Edges"
)
CURVE_DISTRIBUTE_BETWEEN_UVS_LABEL: str = "Curve Distribute Between UVs"
LOFT_DISTRIBUTE_UVS_BETWEEN_EDGES_OR_UVS_LABEL: str = (
    "Loft Distribute UVs Between Edges or UVs"
)
CREATE_CURVE_FROM_EDGES_LABEL: str = "Create Curve from Edges"
CREATE_UV_CURVE_FROM_EDGES_LABEL: str = "Create Curve from Edges in UV Space"
ABOUT_WINDOW: str = "zenToolsAboutWindow"
CLOSE_CHECKBOX: str = "zenToolsCloseCheckBox"


def show_about() -> None:
    """
    Show a window with information about, and button to update/upgrade or
    uninstall, ZenTools.
    """
    if cmds.window(ABOUT_WINDOW, exists=True):
        cmds.deleteUI(ABOUT_WINDOW)
    cmds.window(
        ABOUT_WINDOW,
        title="About ZenTools",
    )
    column_layout: str = cmds.columnLayout(
        parent=ABOUT_WINDOW,
        margins=15,
    )
    version: str = get_maya_zen_tools_package_info()["version"]
    cmds.text(
        label=(
            f"ZenTools {version} © "
            f"{datetime.now(tz=timezone.utc).year} by David Belais\n"
        ),
        align="left",
        parent=column_layout,
    )
    debugging: bool = get_tool_option(  # type: ignore
        "general", "debugging", False
    )
    row_layout: str
    # If debugging is enabled, or if ZenTools is installed as an editable
    # package, show an option to enable/disable debugging
    with contextlib.suppress(Exception):
        package_info: dict[str, str] = get_maya_zen_tools_package_info()
        if debugging or (
            package_info.get("editable_project_location") is not None
        ):
            cmds.checkBox(
                label="Enable Debugging",
                parent=column_layout,
                value=debugging,  # type: ignore
                onCommand=(
                    "from maya_zen_tools import options\n"
                    "options.set_tool_option("
                    "'general', 'debugging', "
                    "True)\n"
                    "from maya_zen_tools import _utilities\n"
                    "_utilities.reload()"
                ),
                offCommand=(
                    "from maya_zen_tools import options\n"
                    "options.set_tool_option("
                    "'general', 'debugging', "
                    "False)\n"
                    "from maya_zen_tools import _utilities\n"
                    "_utilities.reload()"
                ),
                height=30,
            )
    row_layout = cmds.rowLayout(
        parent=column_layout,
        numberOfColumns=2,
    )
    cmds.button(
        label="Update ZenTools",
        parent=row_layout,
        command=(
            "from maya import cmds\n"
            "from maya_zen_tools import upgrade\nupgrade.main()\n"
            f"cmds.deleteUI('{ABOUT_WINDOW}')"
        ),
    )
    cmds.button(
        label="Uninstall ZenTools",
        command=(
            "from maya import cmds\n"
            "from maya_zen_tools import _ui\n"
            "_ui.show_confirmation_dialogue("
            f'"Are you certain you want to uninstall ZenTools?", '
            f'yes_command="from maya_zen_tools import uninstall\\n'
            'uninstall.main()", title="Uninstall ZenTools?")\n'
            f"cmds.deleteUI('{ABOUT_WINDOW}')"
        ),
    )
    cmds.showWindow(ABOUT_WINDOW)


def create_menu() -> None:
    """
    Create the main ZenTools menu
    """
    if cmds.menu(MENU, exists=True):
        cmds.deleteUI(MENU)
    cmds.menu(
        MENU, label="ZenTools", tearOff=True, visible=True, parent=MAYA_WINDOW
    )
    cmds.menuSet(MENU_SET, addMenu=MENU)
    # Selection
    cmds.menuItem(label="Selection", parent=MENU, divider=True)
    cmds.menuItem(
        label=SELECT_EDGES_BETWEEN_VERTICES_LABEL,
        command=(
            "from maya_zen_tools import loop\n"
            "loop.do_select_edges_between_vertices()"
        ),
        annotation=(
            "Selects an edge path containing the fewest edges necessary to "
            "connect selected vertices."
        ),
        parent=MENU,
    )
    cmds.menuItem(
        optionBox=True,
        command=(
            "from maya_zen_tools import loop\n"
            "loop.show_select_edges_between_vertices_options()"
        ),
        parent=MENU,
    )
    cmds.menuItem(
        label=SELECT_EDGES_BETWEEN_UVS_LABEL,
        command=(
            "from maya_zen_tools import loop\n"
            "loop.do_select_edges_between_uvs()"
        ),
        annotation=(
            "Selects an edge path containing the fewest edges necessary to "
            "connect selected UVs."
        ),
        parent=MENU,
    )
    cmds.menuItem(
        optionBox=True,
        command=(
            "from maya_zen_tools import loop\n"
            "loop.show_select_edges_between_uvs_options()"
        ),
        parent=MENU,
    )
    cmds.menuItem(
        label=SELECT_UVS_BETWEEN_UVS_LABEL,
        command=(
            "from maya_zen_tools import loop\n" "loop.do_select_between_uvs()"
        ),
        annotation=(
            "Selects path containing the fewest UVs necessary to "
            "connect selected UVs."
        ),
        parent=MENU,
    )
    cmds.menuItem(
        optionBox=True,
        command=(
            "from maya_zen_tools import loop\n"
            "loop.show_select_between_uvs_options()"
        ),
        parent=MENU,
    )
    cmds.menuItem(
        label="Flood Select",
        command="from maya_zen_tools import flood\nflood.flood_select()",
        annotation=(
            "Selected Edges will define a selection border, selected vertices "
            "or faces will determine the portion of the mesh to be selected."
        ),
        parent=MENU,
    )
    # Modeling
    cmds.menuItem(label="Modeling", parent=MENU, divider=True)
    cmds.menuItem(
        label=CURVE_DISTRIBUTE_BETWEEN_VERTICES_LABEL,
        command=(
            "from maya_zen_tools import loop\n"
            "loop.do_curve_distribute_vertices()"
        ),
        annotation="Align edge loop along a curve based on vertex selection.",
        parent=MENU,
    )
    cmds.menuItem(
        optionBox=True,
        command=(
            "from maya_zen_tools import loop\n"
            "loop.show_curve_distribute_vertices_options()"
        ),
        parent=MENU,
    )
    cmds.menuItem(
        label=LOFT_DISTRIBUTE_VERTICES_BETWEEN_EDGES_LABEL,
        command=(
            "from maya_zen_tools import loft\n"
            "loft.do_loft_distribute_vertices_between_edges()"
        ),
        annotation=(
            "Distribute vertices between two or more parallel edge loops."
        ),
        parent=MENU,
    )
    cmds.menuItem(
        optionBox=True,
        command=(
            "from maya_zen_tools import loft;"
            "loft.show_loft_distribute_vertices_between_edges_options()"
        ),
        parent=MENU,
    )
    # Texturing
    cmds.menuItem(label="Texturing", parent=MENU, divider=True)
    cmds.menuItem(
        label=CURVE_DISTRIBUTE_BETWEEN_UVS_LABEL,
        command=(
            "from maya_zen_tools import loop\n"
            "loop.do_curve_distribute_uvs()"
        ),
        annotation="Align edge loop along a curve based on vertex selection.",
        parent=MENU,
    )
    cmds.menuItem(
        optionBox=True,
        command=(
            "from maya_zen_tools import loop\n"
            "loop.show_curve_distribute_uvs_options()"
        ),
        parent=MENU,
    )
    cmds.menuItem(
        label=LOFT_DISTRIBUTE_UVS_BETWEEN_EDGES_OR_UVS_LABEL,
        command=(
            "from maya_zen_tools import loft\n"
            "loft.do_loft_distribute_uvs_between_edges_or_uvs()"
        ),
        annotation=("Distribute UVs between two or more parallel edge loops."),
        parent=MENU,
    )
    cmds.menuItem(
        optionBox=True,
        command=(
            "from maya_zen_tools import loft;"
            "loft.show_loft_distribute_uvs_between_edges_or_uvs_options()"
        ),
        parent=MENU,
    )
    cmds.menuItem(label="Help", parent=MENU, divider=True)
    cmds.menuItem(
        label="About ZenTools",
        command=(
            "import maya_zen_tools.menu\nmaya_zen_tools.menu.show_about()"
        ),
        parent=MENU,
    )
    cmds.menuItem(
        label="ZenTools Documentation",
        command=(
            "import webbrowser\n"
            "webbrowser.open('https://maya-zen-tools.enorganic.org')"
        ),
        parent=MENU,
    )
    cmds.menuItem(
        label="Reset ZenTools Options",
        command=(
            "import maya_zen_tools.options\nmaya_zen_tools.options.reset()"
        ),
        parent=MENU,
    )
    if get_tool_option("general", "debugging", False):
        # Only show these menu items if `maya-zen-tools` is an
        # editable installation (indicating it is installed for
        cmds.menuItem(label="Debugging", parent=MENU, divider=True)
        # development/testing)
        cmds.menuItem(
            label=CREATE_CURVE_FROM_EDGES_LABEL,
            command=(
                "from maya_zen_tools import loop\n"
                "loop.create_curve_from_edges()"
            ),
            annotation="Create a curve from a contiguous edge selection.",
            parent=MENU,
        )
        cmds.menuItem(
            label=CREATE_UV_CURVE_FROM_EDGES_LABEL,
            command=(
                "from maya_zen_tools import loop\n"
                "loop.create_uv_curve_from_edges()"
            ),
            annotation=(
                "Create a curve from a contiguous edge selection in UV "
                "space"
            ),
            parent=MENU,
        )
        cmds.menuItem(
            label="Reload ZenTools",
            command=(
                "from maya_zen_tools import _utilities\n_utilities.reload()"
            ),
            parent=MENU,
        )



from __future__ import annotations

from typing import Iterable

from maya import cmds  # type: ignore

from maya_zen_tools._traverse import (
    get_component_id,
    get_components_shape,
    get_transform_shape,
    iter_edges_uvs,
    iter_edges_vertices,
)


def create_locator(
    *,
    translate: tuple[float, float, float] | None = None,
    scale: float = 1.0,
    connect_translate: str | tuple[str] | None = None,
    parent: str = "",
) -> str:
    """
    Create a locator at the given translation.

    Parameters:
        translation: The translation of the locator.
        scale: The scale of the locator.
    """
    locator: str = cmds.spaceLocator(name="wireLocator#")[0]
    cmds.setAttr(f"{locator}.translate", *translate)
    locator_shape: str = cmds.listRelatives(
        locator, shapes=True, noIntermediate=True
    )[0]
    cmds.setAttr(f"{locator_shape}.localScale", scale, scale, scale)
    if connect_translate is not None:
        if isinstance(connect_translate, str):
            connect_translate = (connect_translate,)
        connect_translate_to: str
        for connect_translate_to in connect_translate:
            cmds.connectAttr(
                f"{locator}.translate",
                connect_translate_to,
            )
    if parent:
        cmds.parent(locator, parent)
    return locator


def create_edges_rebuild_curve(edges: Iterable[str]) -> str:
    """
    Create a rebuildCurve node from contiguous edges
    """
    edges = tuple(edges)
    vertices: tuple[str, ...] = tuple(iter_edges_vertices(edges))
    polymesh_shape: str = get_components_shape(edges)
    point_on_curve_info: str = cmds.createNode("pointOnCurveInfo")
    cmds.setAttr(f"{point_on_curve_info}.parameter", 0)
    cmds.setAttr(f"{point_on_curve_info}.turnOnPercentage", 1)
    edge: str
    curve_from_mesh_edges: str = ""
    vertex: str
    reverse_previous: bool = True
    reverse: bool = False
    for vertex, edge in zip(vertices[:-1], edges):
        curve_from_mesh_edge: str = cmds.createNode("curveFromMeshEdge")
        cmds.connectAttr(
            f"{polymesh_shape}.worldMesh[0]",
            f"{curve_from_mesh_edge}.inputMesh",
        )
        cmds.setAttr(
            f"{curve_from_mesh_edge}.edgeIndex[0]", get_component_id(edge)
        )
        cmds.connectAttr(
            f"{curve_from_mesh_edge}.outputCurve",
            f"{point_on_curve_info}.inputCurve",
            # Force the connection (replace the pre-existing connection)
            force=True,
        )
        # Determine if we need to reverse the curve
        vertex_point_position: tuple[float, float, float] = tuple(
            cmds.pointPosition(vertex)
        )
        curve_start_point_position: tuple[float, float, float] = cmds.getAttr(
            f"{point_on_curve_info}.position"
        )[0]
        reverse = vertex_point_position != curve_start_point_position
        if not curve_from_mesh_edges:
            # This is the first curve, no need to attach anything
            curve_from_mesh_edges = curve_from_mesh_edge
            # There was no attach node in this iteration, so we carry
            # over reversal to the next iteration
            reverse_previous = reverse
            continue
        # Attach the edge curve to the pre-existing curve
        attach_curve: str = cmds.createNode("attachCurve")
        # Set the attach curve node state to "waiting"
        cmds.setAttr(f"{attach_curve}.nodeState", 8)
        cmds.connectAttr(
            f"{curve_from_mesh_edges}.outputCurve",
            f"{attach_curve}.inputCurve1",
        )
        cmds.connectAttr(
            f"{curve_from_mesh_edge}.outputCurve",
            f"{attach_curve}.inputCurve2",
        )
        # Set the attach curve node state to "active"
        cmds.setAttr(f"{attach_curve}.nodeState", 0)
        # Reverse the curve and/or previous curve so that the start of the
        # curve aligns with `vertex`
        if reverse_previous:
            cmds.setAttr(f"{attach_curve}.reverse1", 1)
        if reverse:
            cmds.setAttr(f"{attach_curve}.reverse2", 1)
        reverse_previous = False
        curve_from_mesh_edges = attach_curve
    rebuild_curve: str = cmds.createNode("rebuildCurve")
    cmds.connectAttr(
        f"{curve_from_mesh_edges}.outputCurve", f"{rebuild_curve}.inputCurve"
    )
    cmds.setAttr(f"{rebuild_curve}.keepControlPoints", 1)
    cmds.setAttr(f"{rebuild_curve}.degree", 1)
    cmds.setAttr(f"{rebuild_curve}.rebuildType", 0)
    cmds.setAttr(f"{rebuild_curve}.spans", len(edges))
    cmds.setAttr(f"{rebuild_curve}.endKnots", 1)
    # Make the range 0 -> # spans
    cmds.setAttr(f"{rebuild_curve}.keepRange", 2)
    return rebuild_curve


def create_uv_edges_rebuild_curve(
    edges: Iterable[str],
) -> tuple[str, str, str]:
    """
    Create a (rebuilt) curve in UV space from contiguous edges
    """
    return create_uvs_rebuild_curve(iter_edges_uvs(edges))


def create_uvs_rebuild_curve(
    uvs: Iterable[str],
) -> tuple[str, str, str]:
    """
    Create a (rebuilt) curve in UV space from contiguous edges
    """
    uvs = tuple(uvs)
    curve_transform: str = cmds.curve(
        editPoint=tuple((*cmds.polyEditUV(uv, query=True), 0.0) for uv in uvs),
        degree=1,
    )
    curve_shape: str = get_transform_shape(curve_transform)
    rebuild_curve: str = cmds.createNode("rebuildCurve")
    cmds.connectAttr(f"{curve_shape}.local", f"{rebuild_curve}.inputCurve")
    cmds.setAttr(f"{rebuild_curve}.keepControlPoints", 1)
    cmds.setAttr(f"{rebuild_curve}.degree", 1)
    cmds.setAttr(f"{rebuild_curve}.rebuildType", 0)
    cmds.setAttr(f"{rebuild_curve}.spans", len(uvs) - 1)
    cmds.setAttr(f"{rebuild_curve}.endKnots", 1)
    # Make the range 0 -> # spans
    cmds.setAttr(f"{rebuild_curve}.keepRange", 2)
    return rebuild_curve, curve_shape, curve_transform



"""
This module modifies your userSetup.py script to add startup procedures
needed to use ZenTools.
"""

from __future__ import annotations

import argparse
import re
from typing import TYPE_CHECKING

from maya_zen_tools._utilities import find_user_setup_py

if TYPE_CHECKING:
    from pathlib import Path


def install() -> None:
    """
    Add the line "from maya_zen_tools import startup" to userSetup.py,
    if it isn't already in the script.
    """
    user_setup_py: str = ""
    user_setup_py_path: Path = find_user_setup_py()
    if user_setup_py_path.is_file():
        with open(user_setup_py_path) as user_setup_py_io:
            user_setup_py = user_setup_py_io.read()
    if not (
        user_setup_py
        and re.search(
            r"(^|\n)from maya_zen_tools import startup(\n|$)", user_setup_py
        )
    ):
        with open(user_setup_py_path, "a") as user_setup_py_io:
            user_setup_py_io.write("from maya_zen_tools import startup\n")


def main() -> None:
    parser: argparse.ArgumentParser = argparse.ArgumentParser(
        prog="maya-zen-tools install",
        description="Install ZenTools for Maya",
    )
    parser.parse_args()
    install()


if __name__ == "__main__":
    main()



from __future__ import annotations

import contextlib
from functools import partial
from math import ceil
from operator import itemgetter
from typing import Callable, Iterable, Sequence

from maya import cmds  # type: ignore

from maya_zen_tools import options
from maya_zen_tools._create import (
    create_edges_rebuild_curve,
    create_locator,
    create_uv_edges_rebuild_curve,
)
from maya_zen_tools._traverse import (
    get_components_shape,
    iter_contiguous_edges,
    iter_contiguous_uv_edges,
    iter_selected_components,
    iter_shortest_uvs_path,
    iter_shortest_uvs_path_proportional_positions,
    iter_shortest_uvs_path_uniform_positions,
    iter_shortest_vertices_path,
    iter_shortest_vertices_path_proportional_positions,
    iter_shortest_vertices_path_uniform_positions,
    iter_sorted_uvs,
    iter_sorted_vertices,
    iter_uvs_edges,
    iter_vertices_edges,
)
from maya_zen_tools._ui import WINDOW
from maya_zen_tools._utilities import as_tuple
from maya_zen_tools.menu import (
    CLOSE_CHECKBOX,
    CURVE_DISTRIBUTE_BETWEEN_UVS_LABEL,
    CURVE_DISTRIBUTE_BETWEEN_VERTICES_LABEL,
    SELECT_EDGES_BETWEEN_UVS_LABEL,
    SELECT_EDGES_BETWEEN_VERTICES_LABEL,
    SELECT_UVS_BETWEEN_UVS_LABEL,
)


def _get_vertices_locator_scale(vertices: Sequence[str]) -> float:
    """
    Get a locator scale appropriate for the given vertices.

    Parameters:
        vertices: A list of vertices.
    """
    bounding_box: tuple[float, float, float, float, float, float] = tuple(
        cmds.exactWorldBoundingBox(*vertices)
    )
    edges: tuple[str] = cmds.ls(
        *cmds.polyListComponentConversion(*vertices, toEdge=True), flatten=True
    )
    average_edge_length: float = sum(map(cmds.arclen, edges)) / len(edges)
    return max(
        average_edge_length / 3,
        max(
            abs(bounding_box[3] - bounding_box[0]),
            abs(bounding_box[4] - bounding_box[1]),
            abs(bounding_box[5] - bounding_box[2]),
        )
        / 10,
    )


def _create_curve_from_vertices(
    vertices: Sequence[str],
    *,
    create_locators: bool = False,
    close: bool = False,
) -> tuple[str, ...]:
    """
    Given a selection of vertices along a shared edge loop, create a curve
    passing between the vertices.

    The curve type will be an *arc* if 3 vertices are selected, otherwise the
    curve will emulate an "edit point" (EP) curve by creating a curve from
    a 0-width loft.

    Parameters:
        vertices: A list of vertices to create the curve from.
        create_locators: If `True`, create locators for manipulating the curve.
        close: If `True`, the curve will form a closed loop.
    """
    curve_transform: str = cmds.createNode(
        "transform", name="wire#", skipSelect=True
    )
    curve_shape: str = cmds.createNode(
        "nurbsCurve",
        name=f"{curve_transform}Shape",
        parent=curve_transform,
        skipSelect=True,
    )
    locator_scale: float = _get_vertices_locator_scale(vertices)
    index: int
    translation: tuple[float, float, float]
    locators: list[str] = []
    if len(vertices) == 3 and not close:  # noqa: PLR2004
        arc: str = cmds.createNode("makeThreePointCircularArc")
        for index, vertex in enumerate(vertices, 1):
            translation = cmds.xform(
                vertex, query=True, worldSpace=True, translation=True
            )
            if create_locators:
                locators.append(
                    create_locator(
                        translate=translation,
                        scale=locator_scale,
                        connect_translate=f"{arc}.point{index}",
                        parent=curve_transform,
                    )
                )
            else:
                cmds.setAttr(
                    f"{arc}.point{index}",
                    *translation,
                )
        cmds.connectAttr(f"{arc}.outputCurve", f"{curve_shape}.create")
    else:
        loft: str = cmds.createNode("loft")
        curve_from_surface_iso: str = cmds.createNode("curveFromSurfaceIso")
        cmds.setAttr(f"{curve_from_surface_iso}.isoparmDirection", 0)
        cmds.setAttr(f"{loft}.uniform", 0)
        if close:
            cmds.setAttr(f"{loft}.close", 1)
        for index, vertex in enumerate(vertices, 0):
            translation = cmds.xform(
                vertex, query=True, worldSpace=True, translation=True
            )
            point_matrix_mult: str = cmds.createNode("pointMatrixMult")
            if create_locators:
                locators.append(
                    create_locator(
                        translate=translation,
                        scale=locator_scale,
                        connect_translate=(f"{point_matrix_mult}.inPoint",),
                    )
                )
            else:
                cmds.setAttr(f"{point_matrix_mult}.inPoint", *translation)
            # Create a 0-length curve to use as an edit point in a loft
            # curve, parent that curve under our output curve transform
            # node, and delete the loft curve's original transform node
            loft_curve_transform: str = cmds.curve(
                objectSpace=True,
                degree=1,
                point=((0, 0, 0), (0, 0, 0)),
            )
            loft_curve_shape: str = cmds.listRelatives(
                loft_curve_transform, shapes=True, noIntermediate=True
            )[0]
            cmds.parent(
                loft_curve_shape, curve_transform, addObject=True, shape=True
            )
            cmds.delete(loft_curve_transform)
            # Connect the inverse matrix from our transform node to the
            # in-matrix of the point matrix multipliers
            cmds.connectAttr(
                f"{curve_transform}.worldInverseMatrix",
                f"{point_matrix_mult}.inMatrix",
            )
            # Connect the point matrix multiplier to the loft curve control
            # points
            cmds.connectAttr(
                f"{point_matrix_mult}.output",
                f"{loft_curve_shape}.controlPoints[0]",
            )
            cmds.connectAttr(
                f"{point_matrix_mult}.output",
                f"{loft_curve_shape}.controlPoints[1]",
            )
            # Connect the loft to the output curve shape
            cmds.connectAttr(
                f"{loft_curve_shape}.worldSpace[0]",
                f"{loft}.inputCurve[{index}]",
            )
        cmds.connectAttr(
            f"{loft}.outputSurface", f"{curve_from_surface_iso}.inputSurface"
        )
        cmds.connectAttr(
            f"{curve_from_surface_iso}.outputCurve", f"{curve_shape}.create"
        )
    return (curve_transform, curve_shape, *locators)


def _create_curve_from_uvs(
    uvs: Sequence[str],
    *,
    close: bool = False,
) -> tuple[str, ...]:
    """
    Given a selection of UVs along a shared edge loop, create a curve
    passing between the UVs.

    The curve type will be an *arc* if 3 UVs are selected, otherwise the
    curve will emulate an "edit point" (EP) curve by creating a curve from
    a 0-width loft.

    Parameters:
        uvs: A list of UVs to create the curve from.
        close: If `True`, the curve will form a closed loop.
    """
    transform: str = cmds.createNode(
        "transform", name="zenLoopCurve#", skipSelect=True
    )
    shape: str = cmds.createNode(
        "nurbsCurve",
        name="zenLoopCurveShape#",
        parent=transform,
        skipSelect=True,
    )
    index: int
    translation: tuple[float, float, float]
    if len(uvs) == 3 and not close:  # noqa: PLR2004
        arc: str = cmds.createNode("makeThreePointCircularArc")
        for index, uv in enumerate(uvs, 1):
            translation = (*cmds.polyEditUV(uv, query=True), 0)
            cmds.setAttr(
                f"{arc}.point{index}",
                *translation,
            )
        cmds.connectAttr(f"{arc}.outputCurve", f"{shape}.create")
    else:
        loft: str = cmds.createNode("loft")
        curve_from_surface_iso: str = cmds.createNode("curveFromSurfaceIso")
        cmds.setAttr(f"{curve_from_surface_iso}.isoparmDirection", 0)
        cmds.setAttr(f"{loft}.uniform", 0)
        if close:
            cmds.setAttr(f"{loft}.close", 1)
        for index, uv in enumerate(uvs, 0):
            translation = (*cmds.polyEditUV(uv, query=True), 0)
            point_matrix_mult: str = cmds.createNode("pointMatrixMult")
            cmds.setAttr(f"{point_matrix_mult}.inPoint", *translation)
            # Create a 0-length curve to use as an edit point in a loft
            # curve, parent that curve under our output curve transform
            # node, and delete the loft curve's original transform node
            loft_curve_transform: str = cmds.curve(
                objectSpace=True,
                degree=1,
                point=((0, 0, 0), (0, 0, 0)),
            )
            loft_curve_shape: str = cmds.listRelatives(
                loft_curve_transform, shapes=True, noIntermediate=True
            )[0]
            cmds.parent(
                loft_curve_shape, transform, addObject=True, shape=True
            )
            cmds.delete(loft_curve_transform)
            # Connect the inverse matrix from our transform node to the
            # in-matrix of the point matrix multipliers
            cmds.connectAttr(
                f"{transform}.worldInverseMatrix",
                f"{point_matrix_mult}.inMatrix",
            )
            # Connect the point matrix multiplier to the loft curve control
            # points
            cmds.connectAttr(
                f"{point_matrix_mult}.output",
                f"{loft_curve_shape}.controlPoints[0]",
            )
            cmds.connectAttr(
                f"{point_matrix_mult}.output",
                f"{loft_curve_shape}.controlPoints[1]",
            )
            # Connect the loft to the output curve shape
            cmds.connectAttr(
                f"{loft_curve_shape}.worldSpace[0]",
                f"{loft}.inputCurve[{index}]",
            )
        cmds.connectAttr(
            f"{loft}.outputSurface", f"{curve_from_surface_iso}.inputSurface"
        )
        cmds.connectAttr(
            f"{curve_from_surface_iso}.outputCurve", f"{shape}.create"
        )
    return (transform, shape)


def _create_wire_deformer(
    deform_curve_attribute: str,
    base_curve_attribute: str,
    vertices: Iterable[str],
) -> str:
    """
    Create a wire deformer to manipulate specified vertices.
    """
    vertices = tuple(vertices)
    wire: str = cmds.wire(
        vertices,
        # after=True,
        dropoffDistance=(0, float("inf")),
    )[0]
    cmds.connectAttr(base_curve_attribute, f"{wire}.baseWire[0]", force=True)
    cmds.connectAttr(
        deform_curve_attribute, f"{wire}.deformedWire[0]", force=True
    )
    return wire


def _distribute_vertices_loop_along_curve(
    selected_vertices: Sequence[str],
    curve_shape: str,
    curve_transform: str,
    *,
    distribution_type: str = options.DistributionType.UNIFORM,
    create_deformer: bool = False,
) -> tuple[str, tuple[str, ...]]:
    """
    Distribute vertices along a curve.

    Parameters:
        selected_vertices: Selected vertices. The distributed vertices
            will be the vertices forming an edge loop between the selected
            vertices.
        curve_shape: The curve shape node.
        distribution_type:
            UNIFORM: Distribute vertices equidistant along the curve.
            PROPORTIONAL: Distribute vertices such that edge lengths are
                proportional to their original lengths in relation the sum
                of all edge lengths.
        sampling: Curve sampling

    Returns:
        A tuple with two items:
        -   The curve shape name. If a deformer is created, this will not be
            the same as the the input curve shape.
        -   A tuple of the vertices distributed, in order
    """
    vertices_positions: tuple[tuple[str, float], ...] = tuple(
        iter_shortest_vertices_path_proportional_positions(selected_vertices)
        if distribution_type == options.DistributionType.PROPORTIONAL
        else iter_shortest_vertices_path_uniform_positions(selected_vertices)
    )
    # Rebuild the curve
    rebuild_curve: str = cmds.createNode("rebuildCurve")
    cmds.connectAttr(
        f"{curve_shape}.worldSpace[0]", f"{rebuild_curve}.inputCurve"
    )
    cmds.setAttr(f"{rebuild_curve}.rebuildType", 0)
    cmds.setAttr(f"{rebuild_curve}.spans", len(selected_vertices) - 1)
    cmds.setAttr(
        f"{rebuild_curve}.degree", cmds.getAttr(f"{curve_shape}.degree")
    )
    cmds.setAttr(f"{rebuild_curve}.keepTangents", 1)
    cmds.setAttr(f"{rebuild_curve}.keepEndPoints", 1)
    cmds.setAttr(f"{rebuild_curve}.keepRange", 2)
    # This point-on-curve info node will slide along the curve to get
    # transform values for the vertices
    point_on_curve_info: str = cmds.createNode("pointOnCurveInfo")
    point_matrix_mult: str = cmds.createNode("pointMatrixMult")
    cmds.connectAttr(
        f"{rebuild_curve}.outputCurve", f"{point_on_curve_info}.inputCurve"
    )
    cmds.connectAttr(
        f"{curve_shape}.worldMatrix[0]", f"{point_matrix_mult}.inMatrix"
    )
    cmds.connectAttr(
        f"{point_on_curve_info}.position", f"{point_matrix_mult}.inPoint"
    )
    vertex: str
    curve_position: float
    for vertex, curve_position in vertices_positions:
        cmds.setAttr(f"{point_on_curve_info}.parameter", curve_position)
        coordinates: tuple[str, str, str] = cmds.getAttr(
            f"{point_matrix_mult}.output"
        )[0]
        cmds.move(*coordinates, vertex, absolute=True, worldSpace=True)
    # Disconnect and delete temporary nodes
    cmds.disconnectAttr(
        f"{rebuild_curve}.outputCurve", f"{point_on_curve_info}.inputCurve"
    )
    cmds.delete(point_on_curve_info)
    cmds.delete(point_matrix_mult)
    if create_deformer:
        rebuilt_curve: str = cmds.createNode(
            "nurbsCurve",
            parent=curve_transform,
        )
        cmds.connectAttr(
            f"{rebuild_curve}.outputCurve", f"{rebuilt_curve}.create"
        )
        cmds.setAttr(f"{curve_shape}.intermediateObject", 1)
        cmds.setAttr(f"{rebuilt_curve}.intermediateObject", 0)
        # Use the rebuilt curve as a wire deformer
        _create_wire_deformer(
            f"{rebuild_curve}.outputCurve",
            f"{rebuilt_curve}.local",
            map(itemgetter(0), vertices_positions),
        )
        cmds.setAttr(f"{curve_shape}.intermediateObject", 0)
        cmds.setAttr(f"{rebuilt_curve}.intermediateObject", 1)
        cmds.setAttr(f"{curve_shape}.visibility", 0)
        # Disconnect the wire base from history, so that changes to the
        # wire aren't negated
        cmds.evalDeferred(
            lambda: cmds.disconnectAttr(
                f"{rebuild_curve}.outputCurve", f"{rebuilt_curve}.create"
            )
        )
        return rebuilt_curve, tuple(map(itemgetter(0), vertices_positions))
    cmds.delete(rebuild_curve)
    return curve_shape, tuple(map(itemgetter(0), vertices_positions))


def _distribute_uvs_loop_along_curve(
    selected_uvs: Sequence[str],
    curve_shape: str,
    *,
    distribution_type: str = options.DistributionType.UNIFORM,
) -> tuple[str, ...]:
    """
    Distribute UVs along a curve.

    Parameters:
        selected_uvs: Selected UVs. The distributed UVs
            will be the UVs forming an edge loop between the selected
            UVs.
        curve_shape: The curve shape node.
        distribution_type:
            UNIFORM: Distribute UVs equidistant along the curve.
            PROPORTIONAL: Distribute UVs such that edge lengths are
                proportional to their original lengths in relation the sum
                of all edge lengths.
        sampling: Curve sampling

    Returns:
        A tuple with two items:
        -   The curve shape name. If a deformer is created, this will not be
            the same as the the input curve shape.
        -   A tuple of the UVs distributed, in order
    """
    uvs_positions: tuple[tuple[str, float], ...] = tuple(
        iter_shortest_uvs_path_proportional_positions(selected_uvs)
        if distribution_type == options.DistributionType.PROPORTIONAL
        else iter_shortest_uvs_path_uniform_positions(selected_uvs)
    )
    # Rebuild the curve
    rebuild_curve: str = cmds.createNode("rebuildCurve")
    cmds.connectAttr(f"{curve_shape}.local", f"{rebuild_curve}.inputCurve")
    cmds.setAttr(f"{rebuild_curve}.rebuildType", 0)
    cmds.setAttr(f"{rebuild_curve}.spans", len(selected_uvs) - 1)
    cmds.setAttr(
        f"{rebuild_curve}.degree", cmds.getAttr(f"{curve_shape}.degree")
    )
    cmds.setAttr(f"{rebuild_curve}.keepTangents", 1)
    cmds.setAttr(f"{rebuild_curve}.keepEndPoints", 1)
    cmds.setAttr(f"{rebuild_curve}.keepRange", 2)
    # This point-on-curve info node will slide along the curve to get
    # transform values for the UVs
    point_on_curve_info: str = cmds.createNode("pointOnCurveInfo")
    point_matrix_mult: str = cmds.createNode("pointMatrixMult")
    cmds.connectAttr(
        f"{rebuild_curve}.outputCurve", f"{point_on_curve_info}.inputCurve"
    )
    cmds.connectAttr(
        f"{curve_shape}.worldMatrix[0]", f"{point_matrix_mult}.inMatrix"
    )
    cmds.connectAttr(
        f"{point_on_curve_info}.position", f"{point_matrix_mult}.inPoint"
    )
    uv: str
    curve_position: float
    for uv, curve_position in uvs_positions:
        cmds.setAttr(f"{point_on_curve_info}.parameter", curve_position)
        position: tuple[str, str, str] = cmds.getAttr(
            f"{point_matrix_mult}.output"
        )[0]
        cmds.polyEditUV(
            uv, uValue=position[0], vValue=position[1], relative=False
        )
    # Disconnect and delete temporary nodes
    cmds.disconnectAttr(
        f"{rebuild_curve}.outputCurve", f"{point_on_curve_info}.inputCurve"
    )
    cmds.delete(point_on_curve_info)
    cmds.delete(point_matrix_mult)
    cmds.delete(rebuild_curve)
    return tuple(map(itemgetter(0), uvs_positions))


def select_edges_between_vertices(
    *selected_vertices: str,
    use_selection_order: bool = False,
    close: bool = False,
) -> tuple[str, ...]:
    """
    Add the edges forming the shortest path between selected vertices to
    the current selection.

    Parameters:
        use_selection_order: If `True`, the edge path will follow the selection
            order, if two or more vertices are selected.
        close: If `True`, the vertices between the last and first selected
            vertex will be included.

    Returns:
        A tuple of the selected edges.
    """
    cmds.waitCursor(state=True)
    try:
        if use_selection_order:
            # Check to make sure that selection order is being tracked, and
            # fall back to automatic sorting if not.
            use_selection_order = cmds.selectPref(
                trackSelectionOrder=True, query=True
            )
        if not use_selection_order:
            close = False
        # If vertices are not explicitly passed, we get them by
        # flattening the current selection of vertices
        selected_vertices = selected_vertices or tuple(
            iter_selected_components("vtx")
        )
        if not use_selection_order:
            # If we have opted not to use selection order, or are unable to
            # because it is not being tracked, we fall back to auomatic sorting
            selected_vertices = tuple(iter_sorted_vertices(selected_vertices))
        edges: tuple[str, ...] = tuple(
            iter_vertices_edges(
                iter_shortest_vertices_path(
                    (*selected_vertices, selected_vertices[0])
                    if close
                    else selected_vertices
                )
            )
        )
        # Select edges
        cmds.select(*edges, add=True)
        # Deselect vertices
        cmds.select(selected_vertices, deselect=True)
    finally:
        if cmds.waitCursor(query=True, state=True):
            cmds.waitCursor(state=False)
    return edges


def select_edges_between_uvs(
    *selected_uvs: str,
    use_selection_order: bool = False,
    close: bool = False,
) -> tuple[str, ...]:
    """
    Add the edges forming the shortest path between selected UVs to
    the current selection.

    Parameters:
        use_selection_order: If `True`, the edge path will follow the selection
            order, if two or more UVs are selected.
        close: If `True`, the UVs between the last and first selected
            UV will be included.

    Returns:
        A tuple of the selected edges.
    """
    cmds.waitCursor(state=True)
    try:
        if use_selection_order:
            # Check to make sure that selection order is being tracked, and
            # fall back to automatic sorting if not.
            use_selection_order = cmds.selectPref(
                trackSelectionOrder=True, query=True
            )
        if not use_selection_order:
            close = False
        # If UVs are not explicitly passed, we get them by
        # flattening the current selection of UVs
        selected_uvs = selected_uvs or tuple(iter_selected_components("map"))
        if not use_selection_order:
            # If we have opted not to use selection order, or are unable to
            # because it is not being tracked, we fall back to auomatic sorting
            selected_uvs = tuple(iter_sorted_uvs(selected_uvs))
        edges: tuple[str, ...] = tuple(
            iter_uvs_edges(
                iter_shortest_uvs_path(
                    (*selected_uvs, selected_uvs[0]) if close else selected_uvs
                )
            )
        )
        # Select edges
        cmds.select(*edges, add=True)
        # Deselect UVs
        cmds.select(selected_uvs, deselect=True)
    finally:
        cmds.waitCursor(state=False)
    return edges


def select_between_uvs(
    *selected_uvs: str,
    use_selection_order: bool = False,
    close: bool = False,
) -> tuple[str, ...]:
    """
    Add the UVs forming the shortest path between selected UVs to
    the current selection.

    Parameters:
        use_selection_order: If `True`, the path will follow the selection
            order, if two or more UVs are selected.
        close: If `True`, the UVs between the last and first selected
            UV will be included.

    Returns:
        A tuple of the selected UVs.
    """
    cmds.waitCursor(state=True)
    try:
        if use_selection_order:
            # Check to make sure that selection order is being tracked, and
            # fall back to automatic sorting if not.
            use_selection_order = cmds.selectPref(
                trackSelectionOrder=True, query=True
            )
        if not use_selection_order:
            close = False
        # If UVs are not explicitly passed, we get them by
        # flattening the current selection of UVs
        selected_uvs = selected_uvs or tuple(iter_selected_components("map"))
        if not use_selection_order:
            # If we have opted not to use selection order, or are unable to
            # because it is not being tracked, we fall back to auomatic sorting
            selected_uvs = tuple(iter_sorted_uvs(selected_uvs))
        uvs: tuple[str, ...] = tuple(
            iter_shortest_uvs_path(
                (*selected_uvs, selected_uvs[0]) if close else selected_uvs
            )
        )
        # Select edges
        cmds.select(*uvs, add=True)
    finally:
        if cmds.waitCursor(query=True, state=True):
            cmds.waitCursor(state=False)
    return uvs


def curve_distribute_vertices(
    *selected_vertices: str,
    distribution_type: str = options.DistributionType.UNIFORM,
    create_deformer: bool = False,
    use_selection_order: bool = False,
    close: bool = False,
) -> tuple[str, ...]:
    """
    Create a curve passing between selected vertices and distribute all
    vertices on the edge loop segment along the curve.

    The curve type will be an *arc* if 3 vertices are selected, otherwise it
    will be an "edit point" (EP) curve.

    Parameters:
        selected_vertices: A list of vertices to create the curve from. If not
            provided, the current selection will be used.
        distribution_type: How to distribute vertices along the curve.
            UNIFORM: Distribute vertices equidistant along the curve.
            PROPORTIONAL: Distribute vertices such that edge lengths are
                proportional to their original lengths in relation the sum
                of all edge lengths.
        create_deformer: If `True`, create a deformer.
        use_selection_order: If `True`, the curve will be created in selection
            order, otherwise, it will be automatically sorted.
        close: If `True`, the curve distribution will form a closed loop, with
            the first selected vertex also being the last.

    Returns:
        A tuple of the affected edges (the same as the end state selection if
        `create_deformer == False`).
    """
    cmds.waitCursor(state=True)
    try:
        if use_selection_order:
            # Check to make sure that selection order is being tracked, and
            # fall back to automatic sorting if not.
            use_selection_order = cmds.selectPref(
                trackSelectionOrder=True, query=True
            )
        if not use_selection_order:
            close = False
        # Store the original selection
        selection: list[str] = cmds.ls(orderedSelection=True, flatten=True)
        # If vertices are not explicitly passed, we get them by
        # flattening the current selection of vertices
        selected_vertices = selected_vertices or tuple(
            iter_selected_components("vtx")
        )
        # Raise an error if selected vertices span more than one mesh
        get_components_shape(selected_vertices)
        if not use_selection_order:
            # If we have opted not to use selection order, or are unable to
            # because it is not being tracked, we fall back to auomatic sorting
            selected_vertices = tuple(iter_sorted_vertices(selected_vertices))
        # Create the Curve
        curve_transform: str
        curve_shape: str
        locators: list[str]
        curve_transform, curve_shape, *locators = _create_curve_from_vertices(
            selected_vertices, create_locators=create_deformer, close=close
        )
        # Distribute Vertices Along the Curve
        vertices: tuple[str, ...]
        curve_shape, vertices = _distribute_vertices_loop_along_curve(
            (
                (*selected_vertices, selected_vertices[0])
                if close
                else selected_vertices
            ),
            curve_shape,
            curve_transform,
            distribution_type=distribution_type,
            create_deformer=create_deformer,
        )
        edges: tuple[str, ...] = tuple(iter_vertices_edges(vertices))
        if not create_deformer:
            # Cleanup the curve and history if not needed for creating a
            # deformer
            cmds.delete(curve_shape, constructionHistory=True)
            cmds.delete(curve_transform, constructionHistory=True)
            cmds.delete(curve_transform)
            cmds.select(*selection)
            cmds.select(*edges, add=True)
            cmds.select(selected_vertices, deselect=True)
            if cmds.waitCursor(query=True, state=True):
                cmds.waitCursor(state=False)
            return edges
        # Go into object selection mode, in order to manipulate locators
        cmds.selectMode(object=True)
        # Select a center locator, if there are more than two, otherwise select
        # an end locator
        cmds.select(locators[ceil(len(locators) / 2) - 1])
    finally:
        if cmds.waitCursor(query=True, state=True):
            cmds.waitCursor(state=False)
    return edges


def curve_distribute_uvs(
    *selected_uvs: str,
    distribution_type: str = options.DistributionType.UNIFORM,
    use_selection_order: bool = False,
    close: bool = False,
) -> tuple[str, ...]:
    """
    Create a curve passing between selected UVs and distribute all
    UVs on the edge loop segment along the curve.

    The curve type will be an *arc* if 3 UVs are selected, otherwise it
    will be an "edit point" (EP) curve.

    Parameters:
        selected_uvs: A list of UVs to create the curve from. If not
            provided, the current selection will be used.
        distribution_type: How to distribute UVs along the curve.
            UNIFORM: Distribute UVs equidistant along the curve.
            PROPORTIONAL: Distribute UVs such that edge lengths are
                proportional to their original lengths in relation the sum
                of all edge lengths.
        use_selection_order: If `True`, the curve will be created in selection
            order, otherwise, it will be automatically sorted.
        close: If `True`, the curve distribution will form a closed loop, with
            the first selected vertex also being the last.

    Returns:
        A tuple of the affected UVs.
    """
    cmds.waitCursor(state=True)
    try:
        if use_selection_order:
            # Check to make sure that selection order is being tracked, and
            # fall back to automatic sorting if not.
            use_selection_order = cmds.selectPref(
                trackSelectionOrder=True, query=True
            )
        if not use_selection_order:
            close = False
        # Store the original selection
        selection: list[str] = cmds.ls(orderedSelection=True, flatten=True)
        # If UVs are not explicitly passed, we get them by
        # flattening the current selection of UVs
        selected_uvs = selected_uvs or tuple(iter_selected_components("map"))
        # Raise an error if selected UVs span more than one mesh
        get_components_shape(selected_uvs)
        if not use_selection_order:
            # If we have opted not to use selection order, or are unable to
            # because it is not being tracked, we fall back to auomatic sorting
            selected_uvs = tuple(iter_sorted_uvs(selected_uvs))
        # Create the Curve
        curve_transform: str
        curve_shape: str
        curve_transform, curve_shape = _create_curve_from_uvs(
            selected_uvs, close=close
        )
        # Distribute UVs Along the Curve
        uvs: tuple[str, ...] = _distribute_uvs_loop_along_curve(
            ((*selected_uvs, selected_uvs[0]) if close else selected_uvs),
            curve_shape,
            distribution_type=distribution_type,
        )
        edges: tuple[str, ...] = tuple(iter_uvs_edges(uvs))
        # Cleanup the curve and history
        cmds.delete(curve_shape, constructionHistory=True)
        cmds.delete(curve_transform, constructionHistory=True)
        cmds.delete(curve_transform)
        cmds.select(*selection, *uvs, add=True)
    finally:
        if cmds.waitCursor(query=True, state=True):
            cmds.waitCursor(state=False)
    return edges


@as_tuple
def create_curve_from_edges(*selected_edges: str) -> Iterable[str]:
    edges: tuple[str, ...]
    selected_edges = selected_edges or tuple(iter_selected_components("e"))
    for edges in iter_contiguous_edges(*selected_edges):
        rebuild_curve: str = create_edges_rebuild_curve(edges)
        curve_transform: str = cmds.createNode(
            "transform", name="curveFromEdges#"
        )
        curve_shape: str = cmds.createNode(
            "nurbsCurve",
            parent=curve_transform,
            name=f"{curve_transform}Shape",
        )
        cmds.connectAttr(
            f"{rebuild_curve}.outputCurve", f"{curve_shape}.create"
        )
        yield curve_shape


@as_tuple
def create_uv_curve_from_edges(*selected_edges: str) -> Iterable[str]:
    edges: tuple[str, ...]
    selected_edges = selected_edges or tuple(iter_selected_components("e"))
    for edges in iter_contiguous_uv_edges(*selected_edges):
        curve_shape: str = create_uv_edges_rebuild_curve(edges)[1]
        yield curve_shape


def show_curve_distribute_vertices_options() -> None:
    """
    Show a window with options to use when executing
    `curve_distribute_vertices`.
    """
    # Get saved options
    get_option: Callable[[str], str | int | float | None] = partial(
        options.get_tool_option, "curve_distribute_vertices"
    )
    # Create the window
    if cmds.window(WINDOW, exists=True):
        cmds.deleteUI(WINDOW)
    cmds.window(
        WINDOW,
        width=240,
        height=100,
        title=f"ZenTools: {CURVE_DISTRIBUTE_BETWEEN_VERTICES_LABEL} Options",
    )
    column_layout: str = cmds.columnLayout(
        adjustableColumn=True, parent=WINDOW, columnAlign="left", margins=15
    )
    selected: int = 1
    with contextlib.suppress(ValueError):
        selected = ("UNIFORM", "PROPORTIONAL").index(
            get_option(  # type: ignore
                "distribution_type", options.DistributionType.UNIFORM
            )
        ) + 1
    cmds.radioButtonGrp(
        label="Distribution Type:",
        parent=column_layout,
        numberOfRadioButtons=2,
        label1="Uniform",
        label2="Proportional",
        columnAlign=(1, "left"),
        changeCommand1=(
            "from maya_zen_tools import options\n"
            "options.set_tool_option("
            "'curve_distribute_vertices', 'distribution_type', "
            "'UNIFORM')"
        ),
        changeCommand2=(
            "from maya_zen_tools import options\n"
            "options.set_tool_option("
            "'curve_distribute_vertices', 'distribution_type', "
            "'PROPORTIONAL')"
        ),
        select=selected,
        height=30,
    )
    cmds.separator(parent=column_layout)
    use_selection_order: bool = get_option(  # type: ignore
        "use_selection_order", False
    )
    cmds.checkBox(
        label="Use Selection Order",
        parent=column_layout,
        value=use_selection_order,  # type: ignore
        onCommand=(
            "from maya_zen_tools import options\n"
            "options.set_tool_option("
            "'curve_distribute_vertices', 'use_selection_order', "
            "True)\n"
            "from maya import cmds\n"
            f"cmds.disable('{CLOSE_CHECKBOX}', value=False)"
        ),
        offCommand=(
            "from maya_zen_tools import options\n"
            "options.set_tool_option("
            "'curve_distribute_vertices', 'use_selection_order', "
            "False)\n"
            "from maya import cmds\n"
            f"cmds.disable('{CLOSE_CHECKBOX}', value=True)"
        ),
        height=30,
    )
    cmds.separator(parent=column_layout)
    cmds.checkBox(
        label="Create Deformer",
        parent=column_layout,
        value=get_option("create_deformer", False),  # type: ignore
        onCommand=(
            "from maya_zen_tools import options\n"
            "options.set_tool_option("
            "'curve_distribute_vertices', 'create_deformer', "
            "True)"
        ),
        offCommand=(
            "from maya_zen_tools import options\n"
            "options.set_tool_option("
            "'curve_distribute_vertices', 'create_deformer', "
            "False)"
        ),
        height=30,
    )
    cmds.separator(parent=column_layout)
    cmds.checkBox(
        CLOSE_CHECKBOX,
        label="Close",
        parent=column_layout,
        value=get_option("close", False),  # type: ignore
        onCommand=(
            "from maya_zen_tools import options\n"
            "options.set_tool_option("
            "'curve_distribute_vertices', 'close', "
            "True)"
        ),
        offCommand=(
            "from maya_zen_tools import options\n"
            "options.set_tool_option("
            "'curve_distribute_vertices', 'close', "
            "False)"
        ),
        height=30,
    )
    if not use_selection_order:
        cmds.disable(CLOSE_CHECKBOX, value=True)
    cmds.button(
        label="Distribute",
        parent=column_layout,
        command=(
            "from maya_zen_tools import loop\n"
            "from maya import cmds\n"
            "loop.do_curve_distribute_vertices()\n"
            f"cmds.deleteUI('{WINDOW}')"
        ),
    )
    cmds.showWindow(WINDOW)


def do_curve_distribute_vertices() -> None:
    """
    Execute `curve_distribute_vertices`, getting arguments from the UI or
    saved options.
    """
    kwargs: dict[str, float | bool | str] = options.get_tool_options(
        "curve_distribute_vertices"
    )
    curve_distribute_vertices(**kwargs)  # type: ignore


def show_curve_distribute_uvs_options() -> None:
    """
    Show a window with options to use when executing
    `curve_distribute_uvs`.
    """
    # Get saved options
    get_option: Callable[[str], str | int | float | None] = partial(
        options.get_tool_option, "curve_distribute_uvs"
    )
    # Create the window
    if cmds.window(WINDOW, exists=True):
        cmds.deleteUI(WINDOW)
    cmds.window(
        WINDOW,
        width=240,
        height=100,
        title=f"ZenTools: {CURVE_DISTRIBUTE_BETWEEN_UVS_LABEL} Options",
    )
    column_layout: str = cmds.columnLayout(
        adjustableColumn=True, parent=WINDOW, columnAlign="left", margins=15
    )
    selected: int = 1
    with contextlib.suppress(ValueError):
        selected = ("UNIFORM", "PROPORTIONAL").index(
            get_option(  # type: ignore
                "distribution_type", options.DistributionType.UNIFORM
            )
        ) + 1
    cmds.radioButtonGrp(
        label="Distribution Type:",
        parent=column_layout,
        numberOfRadioButtons=2,
        label1="Uniform",
        label2="Proportional",
        columnAlign=(1, "left"),
        changeCommand1=(
            "from maya_zen_tools import options\n"
            "options.set_tool_option("
            "'curve_distribute_uvs', 'distribution_type', "
            "'UNIFORM')"
        ),
        changeCommand2=(
            "from maya_zen_tools import options\n"
            "options.set_tool_option("
            "'curve_distribute_uvs', 'distribution_type', "
            "'PROPORTIONAL')"
        ),
        select=selected,
        height=30,
    )
    cmds.separator(parent=column_layout)
    use_selection_order: bool = get_option(  # type: ignore
        "use_selection_order", False
    )
    cmds.checkBox(
        label="Use Selection Order",
        parent=column_layout,
        value=use_selection_order,  # type: ignore
        onCommand=(
            "from maya_zen_tools import options\n"
            "options.set_tool_option("
            "'curve_distribute_uvs', 'use_selection_order', "
            "True)\n"
            "from maya import cmds\n"
            f"cmds.disable('{CLOSE_CHECKBOX}', value=False)"
        ),
        offCommand=(
            "from maya_zen_tools import options\n"
            "options.set_tool_option("
            "'curve_distribute_uvs', 'use_selection_order', "
            "False)\n"
            "from maya import cmds\n"
            f"cmds.disable('{CLOSE_CHECKBOX}', value=True)"
        ),
        height=30,
    )
    cmds.separator(parent=column_layout)
    cmds.checkBox(
        CLOSE_CHECKBOX,
        label="Close",
        parent=column_layout,
        value=get_option("close", False),  # type: ignore
        onCommand=(
            "from maya_zen_tools import options\n"
            "options.set_tool_option("
            "'curve_distribute_uvs', 'close', "
            "True)"
        ),
        offCommand=(
            "from maya_zen_tools import options\n"
            "options.set_tool_option("
            "'curve_distribute_uvs', 'close', "
            "False)"
        ),
        height=30,
    )
    if not use_selection_order:
        cmds.disable(CLOSE_CHECKBOX, value=True)
    cmds.button(
        label="Distribute",
        parent=column_layout,
        command=(
            "from maya_zen_tools import loop\n"
            "from maya import cmds\n"
            "loop.do_curve_distribute_uvs()\n"
            f"cmds.deleteUI('{WINDOW}')"
        ),
    )
    cmds.showWindow(WINDOW)


def do_curve_distribute_uvs() -> None:
    """
    Execute `curve_distribute_uvs`, getting arguments from the UI or
    saved options.
    """
    kwargs: dict[str, float | bool | str] = options.get_tool_options(
        "curve_distribute_uvs"
    )
    curve_distribute_uvs(**kwargs)  # type: ignore


def show_select_edges_between_vertices_options() -> None:
    """
    Show a window with options to use when executing
    `select_edges_between_vertices`.
    """
    # Get saved options
    get_option: Callable[[str], str | int | float | None] = partial(
        options.get_tool_option, "select_edges_between_vertices"
    )
    # Create the window
    if cmds.window(WINDOW, exists=True):
        cmds.deleteUI(WINDOW)
    cmds.window(
        WINDOW,
        width=240,
        height=100,
        title=f"ZenTools: {SELECT_EDGES_BETWEEN_VERTICES_LABEL} Options",
    )
    column_layout: str = cmds.columnLayout(
        adjustableColumn=True, parent=WINDOW, columnAlign="left", margins=15
    )
    use_selection_order: bool = get_option(  # type: ignore
        "use_selection_order", False
    )
    cmds.checkBox(
        label="Use Selection Order",
        parent=column_layout,
        value=use_selection_order,
        onCommand=(
            "from maya_zen_tools import options\n"
            "options.set_tool_option("
            "'select_edges_between_vertices', 'use_selection_order', "
            "True)\n"
            "from maya import cmds\n"
            f"cmds.disable('{CLOSE_CHECKBOX}', value=False)"
        ),
        offCommand=(
            "from maya_zen_tools import options\n"
            "options.set_tool_option("
            "'select_edges_between_vertices', 'use_selection_order', "
            "False)\n"
            "from maya import cmds\n"
            f"cmds.disable('{CLOSE_CHECKBOX}', value=True)"
        ),
        height=30,
    )
    cmds.separator(parent=column_layout)
    cmds.checkBox(
        CLOSE_CHECKBOX,
        label="Close",
        parent=column_layout,
        value=get_option("close", False),  # type: ignore
        onCommand=(
            "from maya_zen_tools import options\n"
            "options.set_tool_option("
            "'select_edges_between_vertices', 'close', "
            "True)"
        ),
        offCommand=(
            "from maya_zen_tools import options\n"
            "options.set_tool_option("
            "'select_edges_between_vertices', 'close', "
            "False)"
        ),
        height=30,
    )
    if not use_selection_order:
        cmds.disable(CLOSE_CHECKBOX, value=True)
    cmds.button(
        label="Select",
        parent=column_layout,
        command=(
            "from maya_zen_tools import loop\n"
            "from maya import cmds\n"
            "loop.do_select_edges_between_vertices()\n"
            f"cmds.deleteUI('{WINDOW}')"
        ),
    )
    cmds.showWindow(WINDOW)


def do_select_edges_between_vertices() -> None:
    """
    Execute `curve_distribute_vertices`, getting arguments from the UI or
    saved options.
    """
    kwargs: dict[str, float | bool | str] = options.get_tool_options(
        "select_edges_between_vertices"
    )
    select_edges_between_vertices(**kwargs)  # type: ignore


def show_select_edges_between_uvs_options() -> None:
    """
    Show a window with options to use when executing
    `select_edges_between_uvs`.
    """
    # Get saved options
    get_option: Callable[[str], str | int | float | None] = partial(
        options.get_tool_option, "select_edges_between_uvs"
    )
    # Create the window
    if cmds.window(WINDOW, exists=True):
        cmds.deleteUI(WINDOW)
    cmds.window(
        WINDOW,
        width=240,
        height=100,
        title=f"ZenTools: {SELECT_EDGES_BETWEEN_UVS_LABEL} Options",
    )
    column_layout: str = cmds.columnLayout(
        adjustableColumn=True, parent=WINDOW, columnAlign="left", margins=15
    )
    use_selection_order: bool = get_option(  # type: ignore
        "use_selection_order", False
    )
    cmds.checkBox(
        label="Use Selection Order",
        parent=column_layout,
        value=use_selection_order,
        onCommand=(
            "from maya_zen_tools import options\n"
            "options.set_tool_option("
            "'select_edges_between_uvs', 'use_selection_order', "
            "True)\n"
            "from maya import cmds\n"
            f"cmds.disable('{CLOSE_CHECKBOX}', value=False)"
        ),
        offCommand=(
            "from maya_zen_tools import options\n"
            "options.set_tool_option("
            "'select_edges_between_uvs', 'use_selection_order', "
            "False)\n"
            "from maya import cmds\n"
            f"cmds.disable('{CLOSE_CHECKBOX}', value=True)"
        ),
        height=30,
    )
    cmds.separator(parent=column_layout)
    cmds.checkBox(
        CLOSE_CHECKBOX,
        label="Close",
        parent=column_layout,
        value=get_option("close", False),  # type: ignore
        onCommand=(
            "from maya_zen_tools import options\n"
            "options.set_tool_option("
            "'select_edges_between_uvs', 'close', "
            "True)"
        ),
        offCommand=(
            "from maya_zen_tools import options\n"
            "options.set_tool_option("
            "'select_edges_between_uvs', 'close', "
            "False)"
        ),
        height=30,
    )
    if not use_selection_order:
        cmds.disable(CLOSE_CHECKBOX, value=True)
    cmds.button(
        label="Select",
        parent=column_layout,
        command=(
            "from maya_zen_tools import loop\n"
            "from maya import cmds\n"
            "loop.do_select_edges_between_uvs()\n"
            f"cmds.deleteUI('{WINDOW}')"
        ),
    )
    cmds.showWindow(WINDOW)


def do_select_edges_between_uvs() -> None:
    """
    Execute `curve_distribute_uvs`, getting arguments from the UI or
    saved options.
    """
    kwargs: dict[str, float | bool | str] = options.get_tool_options(
        "select_edges_between_uvs"
    )
    select_edges_between_uvs(**kwargs)  # type: ignore


def show_select_between_uvs_options() -> None:
    """
    Show a window with options to use when executing
    `select_between_uvs`.
    """
    # Get saved options
    get_option: Callable[[str], str | int | float | None] = partial(
        options.get_tool_option, "select_between_uvs"
    )
    # Create the window
    if cmds.window(WINDOW, exists=True):
        cmds.deleteUI(WINDOW)
    cmds.window(
        WINDOW,
        width=240,
        height=100,
        title=f"ZenTools: {SELECT_UVS_BETWEEN_UVS_LABEL} Options",
    )
    column_layout: str = cmds.columnLayout(
        adjustableColumn=True, parent=WINDOW, columnAlign="left", margins=15
    )
    use_selection_order: bool = get_option(  # type: ignore
        "use_selection_order", False
    )
    cmds.checkBox(
        label="Use Selection Order",
        parent=column_layout,
        value=use_selection_order,
        onCommand=(
            "from maya_zen_tools import options\n"
            "options.set_tool_option("
            "'select_between_uvs', 'use_selection_order', "
            "True)\n"
            "from maya import cmds\n"
            f"cmds.disable('{CLOSE_CHECKBOX}', value=False)"
        ),
        offCommand=(
            "from maya_zen_tools import options\n"
            "options.set_tool_option("
            "'select_between_uvs', 'use_selection_order', "
            "False)\n"
            "from maya import cmds\n"
            f"cmds.disable('{CLOSE_CHECKBOX}', value=True)"
        ),
        height=30,
    )
    cmds.separator(parent=column_layout)
    cmds.checkBox(
        CLOSE_CHECKBOX,
        label="Close",
        parent=column_layout,
        value=get_option("close", False),  # type: ignore
        onCommand=(
            "from maya_zen_tools import options\n"
            "options.set_tool_option("
            "'select_between_uvs', 'close', "
            "True)"
        ),
        offCommand=(
            "from maya_zen_tools import options\n"
            "options.set_tool_option("
            "'select_between_uvs', 'close', "
            "False)"
        ),
        height=30,
    )
    if not use_selection_order:
        cmds.disable(CLOSE_CHECKBOX, value=True)
    cmds.button(
        label="Select",
        parent=column_layout,
        command=(
            "from maya_zen_tools import loop\n"
            "from maya import cmds\n"
            "loop.do_select_between_uvs()\n"
            f"cmds.deleteUI('{WINDOW}')"
        ),
    )
    cmds.showWindow(WINDOW)


def do_select_between_uvs() -> None:
    """
    Execute `curve_distribute_uvs`, getting arguments from the UI or
    saved options.
    """
    kwargs: dict[str, float | bool | str] = options.get_tool_options(
        "select_between_uvs"
    )
    select_between_uvs(**kwargs)  # type: ignore



from __future__ import annotations

import contextlib
import functools
import importlib
import json
import sys
from subprocess import check_output
from traceback import format_exception
from typing import TYPE_CHECKING, Any, Callable, Iterable

if TYPE_CHECKING:
    from types import ModuleType

import os
from pathlib import Path

has_maya_cmds: bool = True
try:
    from maya import cmds  # type: ignore

    has_maya_cmds = hasattr(cmds, "internalVar")
except ImportError:
    has_maya_cmds = False

MAYA_ZEN_TOOLS: str = "maya-zen-tools"


def which_mayapy() -> Path:
    maya_location: str | None = os.environ.get("MAYA_LOCATION")
    if maya_location:
        return Path(maya_location).joinpath("bin", "mayapy")
    return Path(sys.executable)


def reload() -> None:
    """
    Reload all ZenTools modules
    """
    name: str
    module: ModuleType
    for name, module in tuple(sys.modules.items()):
        if name.startswith("maya_zen_tools.") or name == "maya_zen_tools":
            with contextlib.suppress(ModuleNotFoundError):
                importlib.reload(module)


def get_exception_text() -> str:
    """
    When called within an exception, this function returns a text
    representation of the error matching what is found in
    `traceback.print_exception`, but is returned as a string value rather than
    printing.
    """
    return "".join(format_exception(*sys.exc_info()))


def find_user_setup_py() -> Path:
    """
    Find the userSetup.py script
    """
    scripts_directory: Path
    if has_maya_cmds:
        scripts_directory = Path(cmds.internalVar(userScriptDir=True))
    else:
        # If `maya.cmds`` is not available, we have to install
        # using the non-version-specific Maya scripts directory,
        # as we don't know which version to target
        maya_app_dir: str | None = os.environ.get("MAYA_APP_DIR")
        if not maya_app_dir:
            maya_app_dir = (
                os.path.expanduser("~/Library/Preferences/Autodesk/Maya")
                if sys.platform == "darwin"
                else (
                    os.path.expanduser("~/Documents/Maya")
                    if sys.platform.startswith("win")
                    else os.path.expanduser("~/Maya")
                )
            )
        scripts_directory = Path(maya_app_dir) / "scripts"
    os.makedirs(scripts_directory, exist_ok=True)
    return scripts_directory / "userSetup.py"


def get_maya_zen_tools_package_info() -> dict[str, str]:
    package_info: dict[str, str]
    for package_info in json.loads(
        check_output(
            [str(which_mayapy()), "-m", "pip", "list", "--format", "json"],
            text=True,
        ).strip()
    ):
        if package_info["name"] == MAYA_ZEN_TOOLS:
            return package_info
    raise KeyError(MAYA_ZEN_TOOLS)


def as_tuple(
    user_function: Callable[..., Iterable[Any]],
) -> Callable[..., tuple[Any, ...]]:
    """
    This is a decorator which will return an iterable as a tuple.
    """

    def wrapper(*args: Any, **kwargs: Any) -> tuple[Any, ...]:
        return tuple(user_function(*args, **kwargs) or ())

    return functools.update_wrapper(wrapper, user_function)



from __future__ import annotations

from collections import deque
from functools import cache
from itertools import chain, islice
from math import sqrt
from typing import Iterable, Sequence

from maya import cmds  # type: ignore

from maya_zen_tools.errors import (
    InvalidSelectionError,
    NonContiguousMeshSelectionError,
    NonLinearSelectionError,
    TooManyShapesError,
)


def add_shared_vertex_edges(edges: set[str]) -> set[str]:
    """
    Given one or more edges, return these edges, plus all edges
    sharing a vertex with the input edges.
    """
    return set(
        cmds.ls(
            *cmds.polyListComponentConversion(
                *cmds.polyListComponentConversion(
                    *edges, fromEdge=True, toVertex=True
                ),
                fromVertex=True,
                toEdge=True,
            ),
            flatten=True,
        )
    )


def add_shared_uv_edges(edges: set[str]) -> set[str]:
    """
    Given one or more edges, return these edges, plus all edges
    sharing a UV with the input edges.
    """
    return set(
        cmds.ls(
            *cmds.polyListComponentConversion(
                *cmds.polyListComponentConversion(
                    *edges, fromEdge=True, toUV=True
                ),
                fromUV=True,
                toEdge=True,
            ),
            flatten=True,
        )
    )


def get_shared_vertex_edges(edges: set[str]) -> set[str]:
    """
    Given one or more edges, return all edges
    sharing a vertex with the input edges.
    """
    return add_shared_vertex_edges(edges) - edges


def get_shared_uv_edges(edges: set[str]) -> set[str]:
    """
    Given one or more edges, return all edges
    sharing a UV with the input edges.
    """
    return add_shared_uv_edges(edges) - edges


def add_shared_edge_vertices(vertices: set[str]) -> set[str]:
    """
    Given one or more vertices, return these vertices, plus all vertices
    connected by an edge.
    """
    return set(
        cmds.ls(
            *cmds.polyListComponentConversion(
                *cmds.polyListComponentConversion(
                    *vertices, fromVertex=True, toEdge=True
                ),
                fromEdge=True,
                toVertex=True,
            ),
            flatten=True,
        )
    )


def add_shared_face_edge_uvs(uvs: set[str]) -> set[str]:
    """
    Given one or more UVs, return these UVs, plus all UVs
    connected by an edge *and* a face.
    """
    return set(
        cmds.ls(
            *cmds.polyListComponentConversion(
                *cmds.polyListComponentConversion(
                    *uvs, fromUV=True, toEdge=True
                ),
                fromEdge=True,
                toUV=True,
            ),
            flatten=True,
        )
    ) & set(
        cmds.ls(
            *cmds.polyListComponentConversion(
                *cmds.polyListComponentConversion(
                    *uvs, fromUV=True, toFace=True
                ),
                toFace=True,
                toUV=True,
            ),
            flatten=True,
        )
    )


def get_shared_edge_vertices(vertices: set[str]) -> set[str]:
    """
    Given one or more vertices, return all vertices connected to the
    input vertices by an edge.
    """
    return add_shared_edge_vertices(vertices) - vertices


def get_shared_face_edge_uvs(uvs: set[str]) -> set[str]:
    """
    Given one or more UVs, return all UVs connected to the
    input UVs by an edge *and* a face.
    """
    return add_shared_face_edge_uvs(uvs) - uvs


def iter_sort_vertices_by_distance(
    origin_vertex: str, other_vertices: set[str]
) -> Iterable[str]:
    """
    Given an origin vertex and a set of other vertices, yield the other
    vertices by their distance from the origin.

    Parameters:
        origin_vertex: The vertex to use as an origin for sorting.
        other_vertices: The vertices to be sorted.
    """
    vertices: set[str] = {origin_vertex}
    bordering_vertices: set[str]
    matched_vertices: set[str]
    unsorted_vertices: set[str] = set(other_vertices) - {origin_vertex}
    while unsorted_vertices:
        bordering_vertices = get_shared_edge_vertices(vertices)
        if not bordering_vertices:
            # If there are no bordering vertices, any remaining vertices
            # must belong to a part of the mesh which cannot be reached
            # by edge traversal, and is therefore disconnected
            raise NonContiguousMeshSelectionError(
                origin_vertex, other_vertices
            )
        matched_vertices = bordering_vertices & unsorted_vertices
        if matched_vertices:
            yield from matched_vertices
            unsorted_vertices -= matched_vertices
            if not unsorted_vertices:
                return
        # Add the bordering vertices to our traversal selection
        vertices |= bordering_vertices


def iter_sort_uvs_by_distance(
    origin_uv: str, other_uvs: set[str]
) -> Iterable[str]:
    """
    Given an origin UV and a set of other UVs, yield the other
    UVs by their distance from the origin.

    Parameters:
        origin_uv: The vertex to use as an origin for sorting.
        other_uvs: The vertices to be sorted.
    """
    uvs: set[str] = {origin_uv}
    bordering_uvs: set[str]
    matched_uvs: set[str]
    while other_uvs:
        bordering_uvs = get_shared_face_edge_uvs(uvs)
        if not bordering_uvs:
            # If there are no bordering UVs, any remaining UVs
            # must belong to a part of the mesh which cannot be reached
            # by edge traversal, and is therefore disconnected
            raise NonContiguousMeshSelectionError({origin_uv} | other_uvs)
        matched_uvs = bordering_uvs & other_uvs
        if matched_uvs:
            yield from matched_uvs
            other_uvs -= matched_uvs
            if not other_uvs:
                return
        # Add the bordering UVs to our traversal selection
        uvs |= bordering_uvs


def find_end_vertex(vertices: Iterable[str], origin_vertex: str = "") -> str:
    """
    Given a selection of vertices, find *one* of the end vertices
    """
    other_vertices: set[str] = set(vertices)
    if not origin_vertex:
        origin_vertex = other_vertices.pop()
    if len(other_vertices) == 1:
        # If there are fewer than 3 vertices, either will be an end vertex
        return other_vertices.pop()
    if not other_vertices:
        return origin_vertex
    # Given any vertex on an edge loop, the most distance vertex will
    # always be one of the end vertices
    return deque(
        iter_sort_vertices_by_distance(origin_vertex, other_vertices), maxlen=1
    )[-1]


def find_end_uv(uvs: Iterable[str], origin_uv: str = "") -> str:
    """
    Given a selection of UVs, find *one* of the end UVs
    """
    other_uvs: set[str] = set(uvs)
    if not origin_uv:
        origin_uv = other_uvs.pop()
    if len(other_uvs) == 1:
        # If there are fewer than 3 UVs, either will be an end vertex
        return other_uvs.pop()
    if not other_uvs:
        return origin_uv
    # Given any UV on an edge loop, the most distant UV will
    # always be one of the end UVs
    return deque(iter_sort_uvs_by_distance(origin_uv, other_uvs), maxlen=1)[-1]


def iter_sorted_contiguous_vertices(vertices: Iterable[str]) -> Iterable[str]:
    """
    Given a set of vertices along an edge loop, yield the vertices in
    order from one end to the other (which end starts is not guaranteed, so
    this should only be used where the direction does not matter).

    Note: This works with closed loops, whereas `iter_sorted_vertices`
    does not.

    Parameters:
        vertices: A sequence of vertices along an edge loop.
    """
    vertices = set(vertices)
    vertex: str = find_end_vertex(vertices)
    vertices.remove(vertex)
    yield vertex
    while vertices:
        vertex = (get_shared_edge_vertices({vertex}) & vertices).pop()
        vertices.remove(vertex)
        yield vertex


def iter_sorted_contiguous_edges(edges: Iterable[str]) -> Iterable[str]:
    """
    Given a set of contiguous edges, yield the edges in
    order from one end to the other (which end starts is not guaranteed, so
    this should only be used where the direction does not matter).

    Parameters:
        edges: Two or more contiguous edges
    """
    edges = set(edges)
    edge: str
    for edge in iter_vertices_edges(
        iter_sorted_contiguous_vertices(
            cmds.ls(
                *cmds.polyListComponentConversion(
                    *edges, fromEdge=True, toVertex=True
                ),
                flatten=True,
            )
        )
    ):
        yield edge
        edges.remove(edge)
    # If the edges formed a closed loop, there will be one more remaining
    yield from edges


def iter_sorted_contiguous_uvs(uvs: Iterable[str]) -> Iterable[str]:
    """
    Given a set of contiguous UVs along an edge loop, yield the UVs in
    order from one end to the other (which end starts is not guaranteed, so
    this should only be used where the direction does not matter).

    Note: This works with closed loops, whereas `iter_sorted_uvs`
    does not.

    Parameters:
        uvs: A sequence of UVs along an edge loop.
    """
    uvs = set(uvs)
    uv: str = find_end_vertex(uvs)
    uvs.remove(uv)
    yield uv
    while uvs:
        uv = (get_shared_face_edge_uvs({uv}) & uvs).pop()
        uvs.remove(uv)
        yield uv


def iter_sorted_vertices(vertices: Iterable[str]) -> Iterable[str]:
    """
    Given a set of vertices along an edge loop, yield the vertices in
    order from one end to the other (which end starts is not guaranteed, so
    this should only be used where the direction does not matter).

    Parameters:
        vertices: A sequence of vertices along an edge loop.
    """
    other_vertices: set[str] = set(vertices)
    end_vertex: str = find_end_vertex(other_vertices)
    other_vertices.remove(end_vertex)
    yield end_vertex
    yield from iter_sort_vertices_by_distance(end_vertex, other_vertices)


def iter_sorted_uvs(uvs: Iterable[str]) -> Iterable[str]:
    """
    Given a set of UVs along an edge loop, yield the UVs in
    order from one end to the other (which end starts is not guaranteed, so
    this should only be used where the direction does not matter).

    Parameters:
        uvs: A sequence of UVs along an edge loop.
    """
    other_uvs: set[str] = set(uvs)
    end_uv: str = find_end_uv(other_uvs)
    other_uvs.remove(end_uv)
    yield end_uv
    yield from iter_sort_uvs_by_distance(end_uv, other_uvs)


def get_component_id(component: str) -> int:
    """
    Given a component name, return the integer ID.
    """
    return int(component.rpartition("[")[-1].rpartition("]")[0])


def get_component_shape(component: str) -> str | None:
    """
    If `selection` is a polygon component, return the shape,
    otherwise, return `None`.
    """
    shape: str
    component_type: str
    shape, component_type = component.partition("[")[0].rpartition(".")[::2]
    return (
        shape
        if component_type in {"vtx", "map", "e", "vtxFace", "f"}
        else None
    )


def get_components_shape(components: Iterable[str]) -> str:
    """
    Given a set of components, return the shape name, or raise an error
    if there is more than one shape
    """
    shapes: set[str] = set(filter(None, map(get_component_shape, components)))
    if len(shapes) > 1:
        raise TooManyShapesError(tuple(sorted(shapes)))
    if not shapes:
        raise InvalidSelectionError(components)
    return shapes.pop()


def get_shape_transform(shape: str) -> str:
    """
    Get the associated transform node for a shape
    """
    return cmds.listRelatives(shape, parent=True, path=True)[0]


def get_transform_shape(transform: str) -> str:
    """
    Get the first associated shape node for a transorm
    """
    child: str
    return next(
        child
        for child in iter(
            cmds.listRelatives(transform, children=True, path=True),
        )
        if cmds.nodeType(child) != "transform"
    )


def iter_selected_components(
    *component_types: str, selection: Sequence[str] = ()
) -> Iterable[str]:
    """
    Yield selected components, in selection order.

    Parameters:
        component_types: vtx | e | map | vtxFace | f
        selection: A flat selection sequence. If not provided,
            `maya.cmds.ls` will be used.
    """
    component_type: str
    component_types_: set[str] = set(component_types)
    selected: str
    component: str
    selected_object: str
    for selected in selection or cmds.ls(orderedSelection=True, flatten=True):
        selected_object, component = selected.rpartition(".")[::2]
        if not selected_object and component:
            continue
        component_type = component.rpartition("[")[0]
        if component_type and (component_type in component_types_):
            yield selected


def get_uvs_shared_edge(*uvs: str) -> str:
    """
    Get the edge shared by two or more UVs
    """
    uv: str
    shared_edges: set | None = None
    for uv in uvs:
        uv_edges: set[str] = set(
            cmds.ls(
                *cmds.polyListComponentConversion(
                    uv, fromUV=True, toEdge=True
                ),
                flatten=True,
            )
        )
        if shared_edges is None:
            shared_edges = uv_edges
        else:
            shared_edges &= uv_edges
    if not shared_edges:
        raise NonLinearSelectionError(uvs)
    return shared_edges.pop()


def iter_uvs_edges(uvs: Iterable[str]) -> Iterable[str]:
    """
    Yield the edges between a series of ordered UVs, in the same
    order as the UVs
    """
    uvs = iter(uvs)
    try:
        start_uv: str = next(uvs)
    except StopIteration:
        return
    end_uv: str
    for end_uv in uvs:
        yield get_uvs_shared_edge(start_uv, end_uv)
        start_uv = end_uv


def iter_vertices_edges(vertices: Iterable[str]) -> Iterable[str]:
    """
    Yield the edges between a series of ordered vertices, in the same
    order as the vertices
    """
    vertices = iter(vertices)
    try:
        start_vertex: str = next(vertices)
    except StopIteration:
        return
    end_vertex: str
    for end_vertex in vertices:
        yield from cmds.polyListComponentConversion(
            start_vertex,
            end_vertex,
            fromVertex=True,
            toEdge=True,
            internal=True,
        )
        start_vertex = end_vertex


def iter_edges_vertices(edges: Iterable[str]) -> Iterable[str]:
    """
    Yield the vertices between a series of ordered edges, in the same
    order as the edges
    """
    edges = iter(edges)
    previous_vertices: set[str] | None = None
    edge: str
    for edge in edges:
        vertices: set[str] = set(
            cmds.ls(
                *cmds.polyListComponentConversion(
                    edge,
                    fromEdge=True,
                    toVertex=True,
                ),
                flatten=True,
            )
        )
        if previous_vertices is not None:
            yield from previous_vertices - vertices
            yield from previous_vertices & vertices
            previous_vertices = vertices - previous_vertices
        else:
            previous_vertices = vertices
    yield from (previous_vertices or ())


def iter_edges_uvs(edges: Iterable[str]) -> Iterable[str]:
    """
    Yield the UVs between a series of ordered edges, in the same
    order as the edges
    """
    edges = iter(edges)
    previous_uvs: set[str] | None = None
    edge: str
    for edge in edges:
        uvs: set[str] = set(
            cmds.ls(
                *cmds.polyListComponentConversion(
                    edge,
                    fromEdge=True,
                    toUV=True,
                ),
                flatten=True,
            )
        )
        if previous_uvs is not None:
            yield from previous_uvs - uvs
            yield from previous_uvs & uvs
            previous_uvs = uvs - previous_uvs
        else:
            previous_uvs = uvs
    yield from (previous_uvs or ())


def get_distance_between(
    position_a: tuple[float, ...],
    position_b: tuple[float, ...],
    *args: tuple[float, ...],
) -> float:
    """
    Get the (total) distance between two or more 3d or 2d coordinates
    """
    a: float
    b: float
    distance: float = sqrt(
        sum(abs(a - b) ** 2 for a, b in zip(position_a, position_b))
    )
    if args:
        return distance + get_distance_between(position_b, *args)
    return distance


def get_least_deviant_midpoint_vertex(
    start_point_position: tuple[float, float, float],
    end_point_position: tuple[float, float, float],
    midpoint_vertices: Iterable[str],
) -> str:
    """
    Get the vertex with coordinates which deviate least from the line
    connecting a start and end point position.
    """
    least_deviant_vertex: str = ""
    least_deviant_length: float = 0.0
    vertex: str
    for vertex in midpoint_vertices:
        length: float = get_distance_between(
            start_point_position,
            cmds.pointPosition(vertex),
            end_point_position,
        )
        if (not least_deviant_vertex) or (length < least_deviant_length):
            least_deviant_length = length
            least_deviant_vertex = vertex
    return least_deviant_vertex


def get_least_deviant_midpoint_uv(
    start_point_position: tuple[float, float],
    end_point_position: tuple[float, float],
    midpoint_uvs: Iterable[str],
) -> str:
    """
    Get the UV with coordinates which deviate least from the line
    connecting a start and end point positions.
    """
    least_deviant_uv: str = ""
    least_deviant_length: float = 0.0
    uv: str
    for uv in midpoint_uvs:
        length: float = get_distance_between(
            start_point_position,
            tuple(cmds.polyEditUV(uv, query=True)),
            end_point_position,
        )
        if (not least_deviant_uv) or (length < least_deviant_length):
            least_deviant_length = length
            least_deviant_uv = uv
    return least_deviant_uv


def iter_shortest_vertex_path(
    start_vertex: str, end_vertex: str
) -> Iterable[str]:
    """
    Get a the vertex path connected by the fewest possible number of edges by
    intersecting expanding rings of vertices from either end.

    Parameters:
        start_vertex: The vertex at the start of the path.
        end_vertex: The vertex at the end of the path.
    """

    @cache
    def get_start_point_position() -> tuple[float, float, float]:
        return tuple(cmds.pointPosition(start_vertex))

    @cache
    def get_end_point_position() -> tuple[float, float, float]:
        return tuple(cmds.pointPosition(end_vertex))

    start_vertex_rings: list[set[str]] = [{start_vertex}]
    end_vertex_rings: list[set[str]] = [{end_vertex}]
    # Getting the component shape is done early
    # in order to raise an error if the vertices are not on the same shape,
    # but is also used when raising an error, and for use with the polySelect
    # command
    shape: str = get_components_shape((start_vertex, end_vertex))
    vertices: set[str] = {start_vertex}
    expanded_vertices: set[str]
    ring_vertices: set[str]
    # Get a set of rings grown from the start vertex
    while end_vertex not in vertices:
        expanded_vertices = add_shared_edge_vertices(vertices)
        ring_vertices = expanded_vertices - vertices
        if not ring_vertices:
            # If we can't expand any further, and still haven't reached
            # the end vertex, it's not a contiguous mesh
            raise NonContiguousMeshSelectionError(shape)
        start_vertex_rings.append(ring_vertices)
        vertices = expanded_vertices
    vertices = {end_vertex}
    # Get a set of rings grown from the end vertex
    while start_vertex not in vertices:
        # Stop when we've reached the start vertex
        if start_vertex in vertices:
            break
        expanded_vertices = add_shared_edge_vertices(vertices)
        ring_vertices = expanded_vertices - vertices
        end_vertex_rings.append(ring_vertices)
        vertices = expanded_vertices
    # We should now have two sets of vertex rings of equal length
    start_vertex_ring: set[str]
    end_vertex_ring: set[str]
    vertex: str = ""
    for start_vertex_ring, end_vertex_ring in zip(
        start_vertex_rings, reversed(end_vertex_rings)
    ):
        ring_intersection: set[str] = start_vertex_ring & end_vertex_ring
        # There will typically be only one intersecting vertex, however
        # when there is more than one shortest (having the least edges)
        # path between the vertices, we need to make sure that the path
        # we choose is contiguous. When multiple contiguous options for
        # traversal exist, we choose the one which is most nearly aligned
        # with the vector between the start and end vertices.
        if vertex and len(ring_intersection) > 1:
            # Intersect with only the vertices adjacent to the previously
            # yielded vertex
            ring_intersection &= add_shared_edge_vertices({vertex})
        if len(ring_intersection) > 1:
            vertex = get_least_deviant_midpoint_vertex(
                get_start_point_position(),
                get_end_point_position(),
                ring_intersection,
            )
        else:
            vertex = ring_intersection.pop()
        yield vertex


def iter_shortest_uv_path(start_uv: str, end_uv: str) -> Iterable[str]:
    """
    Get a the UV path connected by the fewest possible number of edges by
    intersecting expanding rings of UVs from either end.

    Parameters:
        start_uv: The UV at the start of the path.
        end_uv: The UV at the end of the path.
    """

    @cache
    def get_start_point_position() -> tuple[float, float]:
        return tuple(cmds.polyEditUV(start_uv, query=True))

    @cache
    def get_end_point_position() -> tuple[float, float]:
        return tuple(cmds.polyEditUV(end_uv, query=True))

    start_uv_rings: list[set[str]] = [{start_uv}]
    end_uv_rings: list[set[str]] = [{end_uv}]
    # Getting the component shape is done early
    # in order to raise an error if the UVs are not on the same shape,
    # but is also used when raising an error, and for use with the polySelect
    # command
    shape: str = get_components_shape((start_uv, end_uv))
    uvs: set[str] = {start_uv}
    expanded_uvs: set[str]
    ring_uvs: set[str]
    # Get a set of rings grown from the start UV
    while end_uv not in uvs:
        expanded_uvs = add_shared_face_edge_uvs(uvs)
        ring_uvs = expanded_uvs - uvs
        if not ring_uvs:
            # If we can't expand any further, and still haven't reached
            # the end UV, it's not a contiguous mesh
            raise NonContiguousMeshSelectionError(shape)
        start_uv_rings.append(ring_uvs)
        uvs = expanded_uvs
    uvs = {end_uv}
    # Get a set of rings grown from the end UV
    while start_uv not in uvs:
        # Stop when we've reached the start UV
        if start_uv in uvs:
            break
        expanded_uvs = add_shared_face_edge_uvs(uvs)
        ring_uvs = expanded_uvs - uvs
        end_uv_rings.append(ring_uvs)
        uvs = expanded_uvs
    # We should now have two sets of UV rings of equal length
    start_uv_ring: set[str]
    end_uv_ring: set[str]
    uv: str = ""
    for start_uv_ring, end_uv_ring in zip(
        start_uv_rings, reversed(end_uv_rings)
    ):
        ring_intersection: set[str] = start_uv_ring & end_uv_ring
        # There will typically be only one intersecting UV, however
        # when there is more than one shortest (having the least edges)
        # path between the UVs, we need to make sure that the path
        # we choose is contiguous. When multiple contiguous options for
        # traversal exist, we choose the one which is most nearly aligned
        # with the vector between the start and end UVs.
        if uv and len(ring_intersection) > 1:
            # Intersect with only the UVs adjacent to the previously
            # yielded UV
            ring_intersection &= add_shared_face_edge_uvs({uv})
        if len(ring_intersection) > 1:
            uv = get_least_deviant_midpoint_uv(
                get_start_point_position(),
                get_end_point_position(),
                ring_intersection,
            )
        else:
            uv = ring_intersection.pop()
        yield uv


def iter_shortest_vertices_path(vertices: Iterable[str]) -> Iterable[str]:
    """
    Given two or more vertices, yield the vertices forming the shortest
    path between them.

    Parameters:
        vertices: Two or more vertices.
    """
    vertices = iter(vertices)
    try:
        start_vertex: str = next(vertices)
    except StopIteration:
        return
    is_first: bool = True
    end_vertex: str
    for end_vertex in vertices:
        segment_vertices: Iterable[str] = iter_shortest_vertex_path(
            start_vertex,
            end_vertex,
        )
        yield from (
            segment_vertices
            if is_first
            # Skip the first vertex for segments after the first
            else islice(segment_vertices, 1, None)
        )
        start_vertex = end_vertex
        is_first = False


def iter_shortest_uvs_path(uvs: Iterable[str]) -> Iterable[str]:
    """
    Given two or more UVs, yield the UVs forming the shortest
    path between them.

    Parameters:
        uvs: Two or more UVs.
    """
    uvs = iter(uvs)
    try:
        start_uv: str = next(uvs)
    except StopIteration:
        return
    is_first: bool = True
    end_uv: str
    for end_uv in uvs:
        segment_uvs: Iterable[str] = iter_shortest_uv_path(
            start_uv,
            end_uv,
        )
        yield from (
            segment_uvs
            if is_first
            # Skip the first UV for segments after the first
            else islice(segment_uvs, 1, None)
        )
        start_uv = end_uv
        is_first = False


def iter_vertices_path_proportional_positions(
    vertices: Iterable[str], spans: int = 1
) -> Iterable[tuple[str, float]]:
    """
    Given two or more adjacent vertices, yield tuples with each vertex and a
    number from 0 to `spans` indicating where on the path each vertex should
    be positioned for proportional distribution.

    Parameters:
        vertices: Two or more vertices.

    Yields:
        A tuple containing the vertex name and a number from 0-`spans`
        indicating where on the curve or surface the vertex should be
        positioned.
    """
    vertices = vertices if isinstance(vertices, tuple) else tuple(vertices)
    edge_lengths: list[float] = []
    previous_vertex: str = ""
    vertex: str
    edge_length: float
    for vertex in vertices:
        if previous_vertex:
            try:
                edge: str = cmds.polyListComponentConversion(
                    previous_vertex,
                    vertex,
                    fromVertex=True,
                    toEdge=True,
                    internal=True,
                )[0]
            except IndexError as error:
                raise InvalidSelectionError(vertices) from error
            edge_length = cmds.arclen(edge)
            if not edge_length:
                raise ValueError(edge)
            edge_lengths.append(
                # The length of the preceding edge
                edge_length
            )
        else:
            edge_lengths.append(0)
        previous_vertex = vertex
    total_edge_length: float = sum(edge_lengths)
    traversed_edge_length: float = 0.0
    for vertex, edge_length in zip(vertices, edge_lengths):
        traversed_edge_length += edge_length
        yield vertex, spans * (traversed_edge_length / total_edge_length)


def iter_uvs_path_proportional_positions(
    uvs: Iterable[str], spans: int = 1
) -> Iterable[tuple[str, float]]:
    """
    Given two or more adjacent UVs, yield tuples with each UV and a
    number from 0 to `spans` indicating where on the path each UV should
    be positioned for proportional distribution.

    Parameters:
        uvs: Two or more UVs.

    Yields:
        A tuple containing the UV name and a number from 0-`spans`
        indicating where on the curve or surface the UV should be
        positioned.
    """
    uvs = uvs if isinstance(uvs, tuple) else tuple(uvs)
    edge_lengths: list[float] = []
    previous_uv: str = ""
    uv: str
    edge_length: float
    for uv in uvs:
        if previous_uv:
            try:
                edge: str = get_uvs_shared_edge(previous_uv, uv)
            except IndexError as error:
                raise InvalidSelectionError(uvs) from error
            edge_length = cmds.arclen(edge)
            if not edge_length:
                raise ValueError(edge)
            edge_lengths.append(
                # The length of the preceding edge
                edge_length
            )
        else:
            edge_lengths.append(0)
        previous_uv = uv
    total_edge_length: float = sum(edge_lengths)
    traversed_edge_length: float = 0.0
    for uv, edge_length in zip(uvs, edge_lengths):
        traversed_edge_length += edge_length
        yield uv, spans * (traversed_edge_length / total_edge_length)


def iter_shortest_vertices_path_proportional_positions(
    selected_vertices: Iterable[str],
) -> Iterable[tuple[str, float]]:
    """
    Given two or more vertices, yield the vertices forming the shortest
    path between them, along with a number from 0-1 indicating where on the
    path each vertex should be positioned for proportional distribution.

    Parameters:
        vertices: Two or more vertices.

    Yields:
        A tuple containing the vertex name and a number from 0-1 indicating
        where on the path the vertex should be positioned.
    """
    selected_vertices = tuple(selected_vertices)
    yield from iter_vertices_path_proportional_positions(
        iter_shortest_vertices_path(selected_vertices),
        spans=len(selected_vertices) - 1,
    )


def iter_shortest_uvs_path_proportional_positions(
    selected_uvs: Iterable[str],
) -> Iterable[tuple[str, float]]:
    """
    Given two or more UVs, yield the UVs forming the shortest
    path between them, along with a number from 0-1 indicating where on the
    path each UV should be positioned for proportional distribution.

    Parameters:
        uvs: Two or more UVs.

    Yields:
        A tuple containing the UV name and a number from 0-1 indicating
        where on the path the UV should be positioned.
    """
    selected_uvs = tuple(selected_uvs)
    yield from iter_uvs_path_proportional_positions(
        iter_shortest_uvs_path(selected_uvs),
        spans=len(selected_uvs) - 1,
    )


def iter_vertices_path_uniform_positions(
    vertices: Iterable[str], spans: int = 1
) -> Iterable[tuple[str, float]]:
    """
    Given two or more adjacent vertices, yield each along with a number
    from 0-`spans` indicating where on the curve or surface each vertex should
    be positioned for uniform distribution.

    Parameters:
        vertices: Two or more vertices.

    Yields:
        A tuple containing the vertex name and a number from 0-`spans`
        indicating where on the curve or surface the vertex should be
        positioned.
    """
    vertices = vertices if isinstance(vertices, tuple) else tuple(vertices)
    index: int
    edge_length: int = len(vertices) - 1
    for index, vertex in enumerate(
        vertices,
    ):
        yield vertex, (index / edge_length) * spans


def iter_uvs_path_uniform_positions(
    uvs: Iterable[str], spans: int = 1
) -> Iterable[tuple[str, float]]:
    """
    Given two or more adjacent UVs, yield each along with a number
    from 0-`spans` indicating where on the curve or surface each UV should
    be positioned for uniform distribution.

    Parameters:
        uvs: Two or more UVs.

    Yields:
        A tuple containing the UV name and a number from 0-`spans`
        indicating where on the curve or surface the UV should be
        positioned.
    """
    uvs = uvs if isinstance(uvs, tuple) else tuple(uvs)
    index: int
    edge_length: int = len(uvs) - 1
    for index, uv in enumerate(
        uvs,
    ):
        yield uv, (index / edge_length) * spans


def iter_shortest_vertices_path_uniform_positions(
    selected_vertices: Iterable[str],
) -> Iterable[tuple[str, float]]:
    """
    Given two or more vertices, yield the vertices forming the shortest
    path between them, along with a number from 0-1 indicating where on the
    path each vertex should be positioned for uniform distribution.

    Parameters:
        vertices: Two or more vertices.

    Yields:
        A tuple containing the vertex name and a number from 0-1 indicating
        where on the path the vertex should be positioned.
    """
    selected_vertices = tuple(selected_vertices)
    yield from iter_vertices_path_uniform_positions(
        iter_shortest_vertices_path(selected_vertices),
        spans=len(selected_vertices) - 1,
    )


def iter_shortest_uvs_path_uniform_positions(
    selected_uvs: Iterable[str],
) -> Iterable[tuple[str, float]]:
    """
    Given two or more UVs, yield the UVs forming the shortest
    path between them, along with a number from 0-1 indicating where on the
    path each UV should be positioned for uniform distribution.

    Parameters:
        uvs: Two or more UVs.

    Yields:
        A tuple containing the UV name and a number from 0-1 indicating
        where on the path the UV should be positioned.
    """
    selected_uvs = tuple(selected_uvs)
    yield from iter_uvs_path_uniform_positions(
        iter_shortest_uvs_path(selected_uvs),
        spans=len(selected_uvs) - 1,
    )


def _get_contiguous_edges_terminal_vertices(
    edges: Sequence[str],
) -> tuple[str, str]:
    """
    Get the vertices at the start and end of the given sequence
    of contiguous edges.
    """
    if len(edges) == 1:
        # Both edge vertices are ends
        return tuple(
            cmds.ls(
                *cmds.polyListComponentConversion(
                    *edges, fromEdge=True, toVertex=True
                ),
                flatten=True,
            )[:2]
        )
    return (
        (
            set(
                cmds.ls(
                    *cmds.polyListComponentConversion(
                        edges[0], fromEdge=True, toVertex=True
                    ),
                    flatten=True,
                )
            )
            - set(
                cmds.ls(
                    *cmds.polyListComponentConversion(
                        edges[1], fromEdge=True, toVertex=True
                    ),
                    flatten=True,
                )
            )
        ).pop(),
        (
            set(
                cmds.ls(
                    *cmds.polyListComponentConversion(
                        edges[-1], fromEdge=True, toVertex=True
                    ),
                    flatten=True,
                )
            )
            - set(
                cmds.ls(
                    *cmds.polyListComponentConversion(
                        edges[-2], fromEdge=True, toVertex=True
                    ),
                    flatten=True,
                )
            )
        ).pop(),
    )


def _get_rotated_vertex_loop(
    vertices: tuple[str, ...], terminal_vertex: str
) -> tuple[str, ...]:
    """
    Rotate a (closed) vertex loop so the `terminal_vertex` is the first
    and last vertex in the loop.
    """
    if vertices[0] == terminal_vertex:
        return vertices
    index: int = vertices.index(terminal_vertex)
    return (
        # Vertex loops start and end with the same vertex, so we drop the last
        # item when rotating, then include the terminal vertex at both the
        # start and end
        vertices[index:-1] + vertices[: index + 1]
    )


def _get_rotated_uv_loop(
    uvs: tuple[str, ...], terminal_uv: str
) -> tuple[str, ...]:
    """
    Rotate a (closed) UV loop so the `terminal_uv` is the first
    and last UV in the loop.
    """
    # This logic is identical for UVs and vertices, so we can reuse it
    return _get_rotated_vertex_loop(uvs, terminal_uv)


def _iter_directionally_aligned_vertex_loops(
    vertex_loops: Sequence[tuple[str, ...]],
) -> Iterable[tuple[str, ...]]:
    """
    Given a series of vertex loops, reverse where needed to make them all
    sorted in the same direction
    """
    reference_vertex: str = vertex_loops[0][1]
    yield vertex_loops[0]
    vertex_loop: tuple[str, ...]
    for vertex_loop in vertex_loops[1:]:
        penterminus_vertices: tuple[str, str] = (
            vertex_loop[1],
            vertex_loop[-2],
        )
        if (
            tuple(
                iter_sort_vertices_by_distance(
                    reference_vertex, set(penterminus_vertices)
                )
            )
            == penterminus_vertices
        ):
            yield vertex_loop
        else:
            # Since the next-to-last vertex was closed than the second
            # vertex to the reference, we need to reverse the loop
            yield tuple(reversed(vertex_loop))


def _iter_directionally_aligned_uv_loops(
    uv_loops: Sequence[tuple[str, ...]],
) -> Iterable[tuple[str, ...]]:
    """
    Given a series of UV loops, reverse where needed to make them all
    sorted in the same direction
    """
    reference_uv: str = uv_loops[0][1]
    yield uv_loops[0]
    uv_loop: tuple[str, ...]
    for uv_loop in uv_loops:
        penterminus_uvs: tuple[str, str] = (
            uv_loop[1],
            uv_loop[-2],
        )
        if (
            tuple(
                iter_sort_uvs_by_distance(reference_uv, set(penterminus_uvs))
            )
            == penterminus_uvs
        ):
            yield uv_loop
        else:
            # Since the next-to-last UV was closed than the second
            # UV to the reference, we need to reverse the loop
            yield tuple(reversed(uv_loop))


def _iter_align_closed_loop_contiguous_edges(
    edge_loops: list[tuple[str, ...]],
) -> Iterable[tuple[str, ...]]:
    """
    This function yields sorted and aligned rearrangements of the (closed)
    edge loops provided as input.
    """
    if len(edge_loops) == 1:
        yield from edge_loops
        return
    edge_loop: tuple[str, ...]
    vertex_loops: list[tuple[str, ...]] = [
        tuple(iter_edges_vertices(edge_loop)) for edge_loop in edge_loops
    ]
    # Find a corner vertex by getting the furthest vertex from the first start
    # vertex
    other_vertices: set[str] = set()
    vertex_loop: tuple[str, ...]
    for vertex_loop in vertex_loops[1:]:
        other_vertices |= set(vertex_loop)
    # Finding the vertex the furthest from any in one of our loops would work,
    # but we can marginaslly reduce overhead by use a known endpoint and
    # excluding the vertices in that end points looop
    origin_vertex: str = find_end_vertex(
        other_vertices, origin_vertex=vertex_loops[0][0]
    )
    vertex_loop_sets: tuple[set[str], ...] = tuple(map(set, vertex_loops))
    unused_loop_indices: set[int] = set(range(len(vertex_loops)))
    # Now we will find the closest vertex to the origin on each other loop,
    # and align the vertex loops so that each begins/ends with that vertex
    vertices: set[str] = {origin_vertex}
    sorted_vertex_loops: list[tuple[str, ...]] = []
    while unused_loop_indices:
        index: int
        for index in tuple(unused_loop_indices):
            matched_vertices: set[str] = vertices & vertex_loop_sets[index]
            if matched_vertices:
                sorted_vertex_loops.append(
                    _get_rotated_vertex_loop(
                        vertex_loops[index], matched_vertices.pop()
                    )
                )
                unused_loop_indices.remove(index)
                break
        if unused_loop_indices:
            shared_edge_vertices: set = get_shared_edge_vertices(vertices)
            if not shared_edge_vertices:
                # If there are loops unconsumed, but we can't grow the
                # vertex selection any further, the mesh is likely not
                # contiguous
                raise NonContiguousMeshSelectionError(edge_loops)
            vertices |= shared_edge_vertices
    for vertex_loop in _iter_directionally_aligned_vertex_loops(
        sorted_vertex_loops
    ):
        yield tuple(iter_vertices_edges(vertex_loop))


def _iter_align_closed_loop_contiguous_uvs(
    uv_loops: list[tuple[str, ...]],
) -> Iterable[tuple[str, ...]]:
    """
    This function yields sorted and aligned rearrangements of the (closed)
    UV loops provided as input.
    """
    if len(uv_loops) == 1:
        yield from uv_loops
        return
    # Find a corner UV by getting the furthest UV from the first start
    # UV
    other_uvs: set[str] = set()
    uv_loop: tuple[str, ...]
    for uv_loop in uv_loops[1:]:
        other_uvs |= set(uv_loop)
    # Finding the UV the furthest from any in one of our loops would work,
    # but we can marginaslly reduce overhead by use a known endpoint and
    # excluding the UVs in that end points looop
    origin_uv: str = find_end_uv(other_uvs, origin_uv=uv_loops[0][0])
    uv_loop_sets: tuple[set[str], ...] = tuple(map(set, uv_loops))
    unused_loop_indices: set[int] = set(range(len(uv_loops)))
    # Now we will find the closest UV to the origin on each other loop,
    # and align the UV loops so that each begins/ends with that UV
    uvs: set[str] = {origin_uv}
    sorted_uv_loops: list[tuple[str, ...]] = []
    while unused_loop_indices:
        index: int
        for index in tuple(unused_loop_indices):
            matched_uvs: set[str] = uvs & uv_loop_sets[index]
            if matched_uvs:
                sorted_uv_loops.append(
                    _get_rotated_uv_loop(uv_loops[index], matched_uvs.pop())
                )
                unused_loop_indices.remove(index)
                break
        if unused_loop_indices:
            shared_edge_uvs: set = get_shared_face_edge_uvs(uvs)
            if not shared_edge_uvs:
                # If there are loops unconsumed, but we can't grow the
                # UV selection any further, the mesh is likely not
                # contiguous
                raise NonContiguousMeshSelectionError(uv_loops)
            uvs |= shared_edge_uvs
    for uv_loop in _iter_directionally_aligned_uv_loops(sorted_uv_loops):
        yield uv_loop


def is_closed_edge_loop(edges: tuple[str, ...]) -> bool:
    """
    Given a tuple of (sorted) edges, return `True` if they form
    a closed loop, otherwise return `False`
    """
    return bool(
        set(
            cmds.ls(
                *cmds.polyListComponentConversion(
                    edges[0],
                    toVertex=True,
                    fromEdge=True,
                ),
                flatten=True,
            )
        )
        & set(
            cmds.ls(
                *cmds.polyListComponentConversion(
                    edges[-1],
                    toVertex=True,
                    fromEdge=True,
                ),
                flatten=True,
            )
        )
    )


def iter_aligned_contiguous_edges(  # noqa: C901
    *selected_edges: str,
) -> Iterable[tuple[str, ...]]:
    """
    Given one or more groups of aligned contiguous edges, yield a tuple
    containing each set of contiguous edges
    """
    edge_loop_segments: list[tuple[str, ...]] = list(
        iter_contiguous_edges(*selected_edges)
    )
    if (
        len(edge_loop_segments[0]) > 2  # noqa: PLR2004
        and edge_loop_segments[0][0] != edge_loop_segments[0][-1]
        and is_closed_edge_loop(edge_loop_segments[0])
    ):
        # If the first and last edges share a vertex, the edges form a closed
        # loop, so we require an alternate alignment strategy
        yield from _iter_align_closed_loop_contiguous_edges(edge_loop_segments)
        return
    # Get the start and end vertices for each segment
    origin_vertex: str | None = None
    segment_terminal_vertices: tuple[str, ...]
    start_vertices_edges: dict[str, tuple[str, ...]] = {}
    index: int
    for index, segment_terminal_vertices in enumerate(
        map(_get_contiguous_edges_terminal_vertices, edge_loop_segments)
    ):
        if origin_vertex is None:
            start_vertices_edges[segment_terminal_vertices[0]] = tuple(
                edge_loop_segments[index]
            )
            # This vertex will be used to align subsequent segments
            origin_vertex = segment_terminal_vertices[0]
        else:
            sorted_segment_terminal_vertices: tuple[str, ...] = tuple(
                iter_sort_vertices_by_distance(
                    origin_vertex, set(segment_terminal_vertices)
                )
            )
            start_vertices_edges[sorted_segment_terminal_vertices[0]] = tuple(
                edge_loop_segments[index]
                if (
                    sorted_segment_terminal_vertices
                    == segment_terminal_vertices
                )
                # If the terminal vertices changed order when sorted, the
                # segment was inverted, so we reverse it
                else reversed(edge_loop_segments[index])
            )
    # Since it's now unused, we will re-populate `edge_loop_segments`
    # with sorted edge loops
    edge_loop_segments.clear()
    edge_loop_ids: list[set[int]] = []
    # Sort the edge loops
    start_vertex: str
    for start_vertex in iter_sorted_vertices(start_vertices_edges.keys()):
        edge_loop_segment: tuple[str, ...] = start_vertices_edges[start_vertex]
        edge_loop_segments.append(edge_loop_segment)
        edge_loop_ids.append(set(map(get_component_id, edge_loop_segment)))
    # Trim edges which don't have a corresponding edge (one on the same ring)
    # on all other segments
    shape: str = get_components_shape(chain(*edge_loop_segments))
    length: int = len(edge_loop_segments)
    for index, edge_loop_segment in enumerate(edge_loop_segments):
        start_index: int = 0
        stop_index: int | None = None
        segment_length: int = len(edge_loop_segment)
        segment_index: int
        edge: str
        ring_edge_ids: set[int]
        matched: bool
        other_index: int
        for segment_index, edge in enumerate(edge_loop_segment):
            ring_edge_ids = set(
                cmds.polySelect(shape, edgeRing=get_component_id(edge))
            )
            matched = True
            for other_index in range(length):
                if other_index == index:
                    continue
                if not ring_edge_ids & edge_loop_ids[other_index]:
                    matched = False
                    break
            if matched:
                start_index = segment_index
                break
        for segment_index, edge in enumerate(reversed(edge_loop_segment)):
            ring_edge_ids = set(
                cmds.polySelect(shape, edgeRing=get_component_id(edge))
            )
            matched = True
            for other_index in range(length):
                if other_index == index:
                    continue
                if not ring_edge_ids & edge_loop_ids[other_index]:
                    matched = False
                    break
            if matched:
                stop_index = segment_length - segment_index
                break
        yield edge_loop_segment[start_index:stop_index]


def iter_aligned_contiguous_uvs(
    *selected_uvs: str,
) -> Iterable[tuple[str, ...]]:
    """
    Given one or more groups of UVs which are contiguous in the UV space,
    yield a tuple containing each set of contiguous UVs,
    directionally aligned. If one of the contiguous UV sets is not on the
    same mesh as the rest, it will be discarded/ignored.
    """
    uv_loop_segments: list[tuple[str, ...]] = list(
        iter_contiguous_uvs(*selected_uvs)
    )
    if (
        len(uv_loop_segments[0]) > 2  # noqa: PLR2004
        and uv_loop_segments[0][0] != uv_loop_segments[0][-1]
        and (
            # Check to see if the end UVs are adjacent
            uv_loop_segments[0][-1]
            in get_shared_face_edge_uvs({uv_loop_segments[0][0]})
        )
    ):
        # If UVs form a closed loop, we require an alternate alignment
        # strategy
        yield from _iter_align_closed_loop_contiguous_uvs(uv_loop_segments)
        return
    # Get the start and end uvs for each segment
    uv_loop_segment: tuple[str, ...]
    origin_uv: str | None = None
    segment_terminal_uvs: tuple[str, ...]
    start_uvs_loops: dict[str, tuple[str, ...]] = {}
    for index, segment_terminal_uvs in enumerate(
        (uv_loop_segment[0], uv_loop_segment[-1])
        for uv_loop_segment in uv_loop_segments
    ):
        if origin_uv is None:
            start_uvs_loops[segment_terminal_uvs[0]] = tuple(
                uv_loop_segments[index]
            )
            # This uv will be used to align subsequent segments
            origin_uv = segment_terminal_uvs[0]
        else:
            sorted_segment_terminal_uvs: tuple[str, ...] = tuple(
                iter_sort_uvs_by_distance(origin_uv, set(segment_terminal_uvs))
            )
            start_uvs_loops[sorted_segment_terminal_uvs[0]] = tuple(
                uv_loop_segments[index]
                if (sorted_segment_terminal_uvs == segment_terminal_uvs)
                # If the terminal uvs changed order when sorted, the
                # segment was inverted, so we reverse it
                else reversed(uv_loop_segments[index])
            )
    # Sort the UV loops
    # We re-use `uv_loop_segments` as it's no longer being used
    start_uv: str
    uv_loop_segments = [
        start_uvs_loops[start_uv]
        for start_uv in iter_sorted_uvs(start_uvs_loops.keys())
    ]
    # Not we remove end UVs which are further from other segment
    # ends than another UV in the segment (overhanging segment parts)
    for index, uv_loop_segment in enumerate(uv_loop_segments):
        start_index: int = 0
        stop_index: int = len(uv_loop_segment)
        uv_loop_segment_set: set[str] = set(uv_loop_segment)
        other_index: int
        other_uv_loop_segment: tuple[str, ...]
        for other_index, other_uv_loop_segment in enumerate(uv_loop_segments):
            if other_index == index:
                continue
            # If there is a UV closer to the other segment's first UV
            # than the first UV in this segment, we move the start of this
            # segment to that UV
            start_index = max(
                start_index,
                uv_loop_segment.index(
                    next(
                        iter(
                            iter_sort_uvs_by_distance(
                                other_uv_loop_segment[0], uv_loop_segment_set
                            )
                        )
                    )
                ),
            )
            # If there is a UV closer to the other segment's last UV
            # than the last UV in this segment, we move the end of this segment
            # to that UV
            stop_index = min(
                stop_index,
                uv_loop_segment.index(
                    next(
                        iter(
                            iter_sort_uvs_by_distance(
                                other_uv_loop_segment[-1], uv_loop_segment_set
                            )
                        )
                    )
                )
                + 1,
            )
        yield uv_loop_segment[start_index:stop_index]


def iter_contiguous_edges(  # noqa: C901
    *selected_edges: str,
) -> Iterable[tuple[str, ...]]:
    """
    Yield tuples of contiguous edge loop segments.

    Parameters:
        selected_edges: Two or more edge loop segments comprised of equal
            numbers edges each, with ends alignged along perpendicular
            edge loops forming a rectangular section of a mesh.
    """
    edges: set[str] = set(selected_edges)
    edge_loop_segments: list[list[str]] = []
    edge_loop_segment: list[str]
    # Organize edges into contiguous segments
    while edges:
        edge: str = edges.pop()
        adjacent_edges: set[str] = set(get_shared_vertex_edges({edge}))
        # This is a list of edge loop segment indices where the edge could be
        # appended. Since two segments could be matched for the same edge,
        # we need to retain this as a list.
        found: int | None = None
        index: int
        for index, edge_loop_segment in enumerate(edge_loop_segments):
            if not edge_loop_segment:
                continue
            if edge_loop_segment[0] in adjacent_edges:
                # This edge is adjacent to the first edge in the segment
                if found is None:
                    edge_loop_segment.insert(0, edge)
                    found = index
                    # Keep looking to see if there is a second match
                    continue
                else:
                    if edge == edge_loop_segments[found][-1]:
                        edge_loop_segments[found].extend(edge_loop_segment)
                    else:
                        edge_loop_segments[found] = (
                            list(reversed(edge_loop_segment))
                            + edge_loop_segments[found]
                        )
                    edge_loop_segment.clear()
                    break
            if edge_loop_segment[-1] in adjacent_edges:
                # This edge is adjacent to the last edge in the segment
                if found is None:
                    edge_loop_segment.append(edge)
                    found = index
                    # Keep looking to see if there is a second match
                    continue
                else:
                    if edge == edge_loop_segments[found][-1]:
                        edge_loop_segments[found].extend(
                            reversed(edge_loop_segment)
                        )
                    else:
                        edge_loop_segments[found] = (
                            list(edge_loop_segment) + edge_loop_segments[found]
                        )
                    edge_loop_segment.clear()
                    break
        if found is None:
            # This edge is not adjacent to any segments started thus far
            edge_loop_segments.append([edge])
    # Remove the cleared segments (they've been joined with another)
    yield from map(tuple, filter(None, edge_loop_segments))


def iter_contiguous_uv_edges(  # noqa: C901
    *selected_edges: str,
) -> Iterable[tuple[str, ...]]:
    """
    Yield tuples of edge loop segments which are contiguous in the UV space.

    Parameters:
        selected_edges: Two or more edge loop segments comprised of equal
            numbers edges each, with ends alignged along perpendicular
            edge loops forming a rectangular section of a mesh.
    """
    edges: set[str] = set(selected_edges)
    edge_loop_segments: list[list[str]] = []
    edge_loop_segment: list[str]
    # Organize edges into contiguous segments
    while edges:
        edge: str = edges.pop()
        adjacent_edges: set[str] = set(get_shared_uv_edges({edge}))
        # This is a list of edge loop segment indices where the edge could be
        # appended. Since two segments could be matched for the same edge,
        # we need to retain this as a list.
        found: int | None = None
        index: int
        for index, edge_loop_segment in enumerate(edge_loop_segments):
            if not edge_loop_segment:
                continue
            if edge_loop_segment[0] in adjacent_edges:
                # This edge is adjacent to the first edge in the segment
                if found is None:
                    edge_loop_segment.insert(0, edge)
                    found = index
                    # Keep looking to see if there is a second match
                    continue
                else:
                    if edge == edge_loop_segments[found][-1]:
                        edge_loop_segments[found].extend(edge_loop_segment)
                    else:
                        edge_loop_segments[found] = (
                            list(reversed(edge_loop_segment))
                            + edge_loop_segments[found]
                        )
                    edge_loop_segment.clear()
                    break
            if edge_loop_segment[-1] in adjacent_edges:
                # This edge is adjacent to the last edge in the segment
                if found is None:
                    edge_loop_segment.append(edge)
                    found = index
                    # Keep looking to see if there is a second match
                    continue
                else:
                    if edge == edge_loop_segments[found][-1]:
                        edge_loop_segments[found].extend(
                            reversed(edge_loop_segment)
                        )
                    else:
                        edge_loop_segments[found] = (
                            list(edge_loop_segment) + edge_loop_segments[found]
                        )
                    edge_loop_segment.clear()
                    break
        if found is None:
            # This edge is not adjacent to any segments started thus far
            edge_loop_segments.append([edge])
    # Remove the cleared segments (they've been joined with another)
    yield from map(tuple, filter(None, edge_loop_segments))


def iter_contiguous_uvs(  # noqa: C901
    *selected_uvs: str,
) -> Iterable[tuple[str, ...]]:
    """
    Yield tuples of UV loop segment, each of which are contiguous, and all
    of which reside on the same UV shell.

    Parameters:
        selected_uvs: Two or more UV loop segments comprised of equal
            numbers of UVs each, with ends alignged along perpendicular
            edge loops forming a rectangular section of a UV mesh.
    """
    uvs: set[str] = set(selected_uvs)
    uv_loop_segments: list[list[str]] = []
    uv_loop_segment: list[str]
    # This list stores one UV ID per/loop, in order to subsequently determine
    # if any don't share a UV shell
    representative_uv_ids: list[int | None] = []
    # Organize edges into contiguous segments
    while uvs:
        uv: str = uvs.pop()
        adjacent_uvs: set[str] = get_shared_face_edge_uvs({uv})
        # This is a list of UV loop segment indices where the UV could be
        # appended. Since two segments could be matched for the same UV,
        # we need to retain this as a list.
        found: int | None = None
        index: int
        for index, uv_loop_segment in enumerate(uv_loop_segments):
            if not uv_loop_segment:
                continue
            if uv_loop_segment[0] in adjacent_uvs:
                # This edge is adjacent to the first edge in the segment
                if found is None:
                    uv_loop_segment.insert(0, uv)
                    found = index
                    # Keep looking to see if there is a second match
                    continue
                else:
                    if uv == uv_loop_segments[found][-1]:
                        uv_loop_segments[found].extend(uv_loop_segment)
                    else:
                        uv_loop_segments[found] = (
                            list(reversed(uv_loop_segment))
                            + uv_loop_segments[found]
                        )
                    uv_loop_segment.clear()
                    representative_uv_ids[index] = None
                    break
            if uv_loop_segment[-1] in adjacent_uvs:
                # This UV is adjacent to the last UV in the segment
                if found is None:
                    uv_loop_segment.append(uv)
                    found = index
                    # Keep looking to see if there is a second match
                    continue
                else:
                    if uv == uv_loop_segments[found][-1]:
                        uv_loop_segments[found].extend(
                            reversed(uv_loop_segment)
                        )
                    else:
                        uv_loop_segments[found] = (
                            list(uv_loop_segment) + uv_loop_segments[found]
                        )
                    uv_loop_segment.clear()
                    representative_uv_ids[index] = None
                    break
        if found is None:
            # This UV is not adjacent to any segments started thus far
            uv_loop_segments.append([uv])
            representative_uv_ids.append(get_component_id(uv))
    # If there are more than 2 UV loop segments, check to see if one is an
    # orphan, and abandon any orphans
    if len(uv_loop_segments) > 2:  # noqa: PLR2004
        shape: str = get_components_shape(selected_uvs)
        # Clear UV loops which aren't on the same shell as the rest
        uv_id: int | None
        other_index: int
        other_uv_id: int | None
        for index, uv_id in enumerate(representative_uv_ids):
            if uv_id is None:
                continue
            found_path: bool = False
            for other_index, other_uv_id in enumerate(representative_uv_ids):
                if other_uv_id is None:
                    continue
                if index == other_index:
                    continue
                if cmds.polySelect(
                    shape, shortestEdgePathUV=(uv_id, other_uv_id), query=True
                ):
                    found_path = True
                    break
            if not found_path:
                uv_loop_segments[index].clear()
    # Remove the cleared segments (they've been joined with another)
    yield from map(tuple, filter(None, uv_loop_segments))


def get_polymesh_shape_uvs_positions(
    shape: str,
) -> dict[int, tuple[float, float]]:
    """
    Intended for testing, this function returns a mapping of UV IDs
    to the UV-space coordinates
    """
    uvs_positions: dict[int, tuple[float, float]] = {}
    uv: str
    for uv in cmds.ls(f"{shape}.map[*]", flatten=True):
        uvs_positions[get_component_id(uv)] = tuple(
            cmds.polyEditUV(uv, query=True)
        )
    return uvs_positions


def get_polymesh_shape_vertices_positions(
    shape: str,
) -> dict[int, tuple[float, float, float]]:
    """
    Intended for testing, this function returns a mapping of vertex IDs
    to their 3d translation coordinates
    """
    vertices_positions: dict[int, tuple[float, float, float]] = {}
    vertex: str
    for vertex in cmds.ls(f"{shape}.vtx[*]", flatten=True):
        vertices_positions[get_component_id(vertex)] = tuple(
            cmds.pointPosition(vertex)
        )
    return vertices_positions


def get_polymesh_shape_changed_vertices_positions(
    shape: str, vertices_positions: dict[int, tuple[float, float, float]]
) -> dict[int, tuple[float, float, float]]:
    """
    Intended for testing, this function returns a mapping of vertex IDs
    to their 3d translation coordinates for vertices which differ
    from those in the input `vertices_positions`.
    """
    changed_vertices_positions: dict[int, tuple[float, float, float]] = {}
    vertex: str
    for vertex in cmds.ls(f"{shape}.vtx[*]", flatten=True):
        vertex_id: int = get_component_id(vertex)
        point_position: tuple[float, float, float] = tuple(
            cmds.pointPosition(vertex)
        )
        if point_position != vertices_positions.get(vertex_id):
            changed_vertices_positions[vertex_id] = point_position
    return changed_vertices_positions


def get_polymesh_shape_changed_uvs_positions(
    shape: str, uvs_positions: dict[int, tuple[float, float]]
) -> dict[int, tuple[float, float]]:
    """
    Intended for testing, this function returns a mapping of UV IDs
    to their UV space coordinates for UVs which differ
    from those in the input `uvs_positions`.
    """
    changed_uvs_positions: dict[int, tuple[float, float]] = {}
    uv: str
    for uv in cmds.ls(f"{shape}.map[*]", flatten=True):
        uv_id: int = get_component_id(uv)
        point_position: tuple[float, float] = tuple(
            cmds.polyEditUV(uv, query=True)
        )
        if point_position != uvs_positions.get(uv_id):
            changed_uvs_positions[uv_id] = point_position
    return changed_uvs_positions



from __future__ import annotations

from maya import cmds  # type: ignore

WINDOW: str = "zenToolsWindow"


def show_confirmation_dialogue(
    label: str,
    yes_command: str,
    cancel_command: str = "",
    title: str = "",
) -> None:
    """
    Show a dialogue asking for confirmation of an operation.

    Parameters:
        text: The text to display
        yes_command: The command to execute if the user clicks "Yes".
        cancel_command: The command to execute if the user clicks "Cancel".
        title: The title for the dialogue window.
    """
    window: str = cmds.window(title=title or label)
    column_layout: str = cmds.columnLayout(parent=window, margins=15)
    cmds.text(
        label=f"{label.rstrip()}\n",
        align="left",
        parent=column_layout,
    )
    row_layout: str = cmds.rowLayout(parent=column_layout, numberOfColumns=2)
    cmds.button(
        label="Yes",
        parent=row_layout,
        command=(
            f"{yes_command}\n"
            "from maya import cmds\n"
            f"cmds.deleteUI('{window}')"
        ),
    )
    cmds.button(
        label="Cancel",
        parent=row_layout,
        command=(
            f"{cancel_command}\n"
            "from maya import cmds\n"
            f"cmds.deleteUI('{window}')"
        ),
    )
    cmds.showWindow(window)



from __future__ import annotations

from maya import cmds  # type: ignore


def center_pivot(
    transform: str,
) -> tuple[float, float, float]:
    center: tuple[float, float, float] = tuple(cmds.objectCenter(transform))
    cmds.setAttr(f"{transform}.rotatePivot", *center)
    cmds.setAttr(f"{transform}.scalePivot", *center)
    return center



"""
This module modifies your userSetup.py script to add startup procedures
needed to use ZenTools.
"""

from __future__ import annotations

import argparse
from subprocess import check_call

from maya_zen_tools._utilities import (
    get_maya_zen_tools_package_info,
    reload,
    which_mayapy,
)


def upgrade() -> None:
    """
    Install the most recent version of ZenTools for Maya
    """
    # Check to see if the currently installed package is editable
    package_info: dict[str, str] = get_maya_zen_tools_package_info()
    if package_info.get("editable_project_location") is not None:
        # Don't upgrade if installed in editable mode (for development)
        return
    check_call(
        [
            str(which_mayapy()),
            "-m",
            "pip",
            "install",
            "--upgrade",
            "--upgrade-strategy",
            "eager",
            "maya-zen-tools",
        ]
    )
    updated_package_info: dict[str, str] = get_maya_zen_tools_package_info()
    if updated_package_info["version"] != package_info["version"]:
        # Reload all maya_zen_tools modules
        reload()
        # Re-install ZenTools for maya
        from maya_zen_tools import install

        install.main()


def main() -> None:
    parser: argparse.ArgumentParser = argparse.ArgumentParser(
        prog="maya-zen-tools upgrade",
        description="Upgrade ZenTools for Maya",
    )
    parser.parse_args()
    upgrade()


if __name__ == "__main__":
    main()
